import warnings

from calendar import monthrange
from collections import namedtuple
from datetime import datetime, timedelta
from typing import Any, List, Tuple


LIMITS = [[0, 59], [0, 23], [1, 31], [1, 12], [0, 6]]
NAMES = ['minute', 'hour', 'day_of_month', 'month', 'day_of_week']
CronParts = namedtuple('CronParts', NAMES)


class CronParserError(Exception):
    """Base exception for all errors during Cron string parsing"""

    pass


class CronSchedule:
    """
    A class representing a Cron schedule. It supports the full standard Cron dialect,
    i.e. minute, hour, day of month, month, day of week with list, range and increment modifiers.
    Name aliases for weekdays or months (e.g. Mon, Tue, ... and Jan, Feb, ...) are not supported

    Attributes
    ----------
    cron_str: :class:`str`
        the string representation of the Cron schedule

    Examples
    --------
    "0 0 * * *": run every day at midnight
    "15 0/4 1 * *": run at 15 minutes past every fourth hour on the first day of each month
    "0 * 14,28 * *": run every hour on the 14th and 28th day of each month
    "1/2 * * * 1-5": run every second hour starting at 1:00 AM on Monday through Friday
    """

    def __init__(self, cron_str: str):
        try:
            entries = CronParts(*cron_str.split())
        except (TypeError, ValueError) as e:
            raise CronParserError(
                'Invalid Cron string. A Cron string must consist of exactly five entries separated by '
                'whitespaces: minute, hour, day of month, month, day of week. Example: `0 0 * * *`'
            ) from e

        allowed_values = []
        try:
            for name, entry, limits in zip(NAMES, entries, LIMITS):
                allowed_values.append(self.__parse_entry(name, entry, limits))
        except ValueError as e:
            raise CronParserError(
                f'Invalid Cron string. {name.title()} is malformed. A Cron string element must either be the '
                'wildcard `*` or contain one or more (comma-separated) entries. An entry must contain a time '
                'indicator or a time range `start-end`, and it may contain an increment-suffix separated by a '
                'slash `/`. Example: `4,12-18/2`'
            ) from e

        self.cron_str = cron_str
        self._entries = entries
        self.allowed_values = CronParts(*allowed_values)

    def __str__(self) -> str:
        return self.cron_str

    def __eq__(self, other: Any):
        # two Cron schedules can be considered equal if they allow the same values
        # i.e. Cron('0 0 * * *') == Cron('0 0 1-31 * *')
        return self.__class__ == other.__class__ and self.allowed_values == other.allowed_values

    @staticmethod
    def __parse_entry(name: str, entry: str, limits: List[int]) -> List[int]:
        """Parse a single entry of the cron string"""

        allowed_values = []
        for part in entry.split(','):
            # split off increment
            if '/' in part:
                part, increment = part.split('/')
                increment = int(increment)
                if increment <= 0:
                    raise CronParserError('Invalid Cron string. Increments must be > 0')
            else:
                increment = None

            if part == '*':  # wildcard
                start, end = limits[0], limits[1]
            elif '-' in part:  # time range
                start, end = part.split('-')
                start, end = int(start), int(end)
            else:  # single value
                start = int(part)
                end = limits[1] if increment else start

            if start < limits[0] or end > limits[1]:
                raise CronParserError(
                    f'Invalid Cron string. {name.title()} is out of bounds: must be <= {limits[1]}, '
                    f'>= {limits[0]}, got {start if start < limits[0] else end}'
                )

            if increment:
                allowed_values.extend(list(range(start, end + 1, increment)))
            else:
                allowed_values.extend(list(range(start, end + 1)))
        return sorted(allowed_values)

    @staticmethod
    def __next_allowed_val(value: int, allowed_values: List[int]) -> Tuple[int, bool]:
        """Get the next allowed value from a list of choices and indicate if the list has overflown,
        i.e. whether the next value needs to be incremented"""

        for val in allowed_values:
            if val >= value:
                return val, False
        return allowed_values[0], True

    def __determine_day(self, reference_date: datetime, dow: int, dow_overflow: bool, dom: int, dom_overflow: bool):
        """Solve the OR-relation between day of month and day of week, returning whichever is smaller"""

        # translate day of week into its equivalent day of month
        month_days = monthrange(reference_date.year, reference_date.month)[1]
        dow = reference_date.day + (7 + dow - reference_date.isoweekday()) % 7 + \
            (7 if dow_overflow and reference_date.isoweekday() == dow else 0)
        dow_overflow = dow > month_days
        dow = (dow - 1) % month_days + 1

        # figure out which one is smaller
        if self._entries.day_of_week != '*' and self._entries.day_of_month != '*':
            if dow_overflow and not dom_overflow:
                return dom, dom_overflow
            elif dom_overflow and not dow_overflow:
                return dow, dow_overflow
            elif dom <= dow:
                return dom, dom_overflow
            else:
                return dow, dow_overflow
        elif self._entries.day_of_week != '*':
            return dow, dow_overflow
        else:
            return dom, dom_overflow

    def next_run_after(self, after: datetime) -> datetime:
        """Calculate the next run time of the Cron schedule after a given reference datetime
        Parameters
        ----------
        after: :class:`datetime.datetime`
            the reference datetime

        Returns
        -------
        The next run time after the reference datetime for the Cron schedule: :class:`datetime.datetime`.
        If the input datetime was timezone-aware, the return will also be. If the input was timezone-naive,
        so will the return be
        """

        # construct timezone-aware representation of after
        tz_naive = not after.tzinfo
        if tz_naive:
            warnings.warn(
                'The input datetime was datetime-naive. Assuming the time zone of your device for processing. '
                'The return value will be datetime-naive again', category=RuntimeWarning
            )
            after = after.replace(tzinfo=datetime.now().astimezone().tzinfo)

        now_parts = CronParts(after.minute, after.hour, after.day, after.month, after.isoweekday())

        # next run's minute
        next_minute, min_overflow = self.__next_allowed_val(now_parts.minute, self.allowed_values.minute)

        # next run's hour
        next_hour, hr_overflow = self.__next_allowed_val(now_parts.hour + (1 if min_overflow else 0),
                                                         self.allowed_values.hour)
        if hr_overflow or next_hour > now_parts.hour:  # we overflowed into the next hour, backtrack
            next_minute = self.allowed_values.minute[0]

        # next run's day
        next_dow, dow_overflow = self.__next_allowed_val(now_parts.day_of_week + (1 if hr_overflow else 0),
                                                         self.allowed_values.day_of_week)
        next_dom, dom_overflow = self.__next_allowed_val(now_parts.day_of_month + (1 if hr_overflow else 0),
                                                         self.allowed_values.day_of_month)
        if next_dom > monthrange(after.year, after.month)[1]:  # current month doesn't allow this day
            next_dom, dom_overflow = self.allowed_values.day_of_month[0], True
        next_day, day_overflow = self.__determine_day(after, next_dow, dow_overflow, next_dom, dom_overflow)
        if day_overflow or next_day > now_parts.day_of_month:  # backtrack
            next_minute = self.allowed_values.minute[0]
            next_hour = self.allowed_values.hour[0]

        # next run's month & year
        next_month, month_overflow = self.__next_allowed_val(now_parts.month + (1 if day_overflow else 0),
                                                             self.allowed_values.month)
        next_year = after.year + 1 if month_overflow else after.year
        if month_overflow or next_month > now_parts.month:  # backtrack
            next_minute = self.allowed_values.minute[0]
            next_hour = self.allowed_values.hour[0]
            next_dom = self.allowed_values.day_of_month[0]
            next_dow = (monthrange(next_year, next_month)[0] + 1) % 7  # ISO weekday of the next month's first day
            ref = (after.replace(day=1) + timedelta(days=32)).replace(day=1)
            next_day, _ = self.__determine_day(ref, next_dow, dow_overflow, next_dom, dom_overflow)

        if tz_naive:
            return datetime(year=next_year, month=next_month, day=next_day, hour=next_hour, minute=next_minute)
        return datetime(year=next_year, month=next_month, day=next_day, hour=next_hour,
                        minute=next_minute, tzinfo=after.tzinfo)

    @property
    def next_run(self) -> datetime:
        """The next run time for this Cron schedule
        Returns
        -------
        the next run time for this Cron schedule: :class:`datetime.datetime`
        NOTE: the return value will always be timezone-AWARE
        """

        return self.next_run_after(datetime.now().astimezone())





from .abc import LeveledUnit
from .miscmodels import TimeDelta, TID
from .enums import Resource, VillageType, ProductionBuildingType, EquipmentRarity

class Hero(LeveledUnit):
    """Represents a Hero object as returned by the API, optionally filled with game data.

    Attributes
    ----------
    name: :class:`str`
        The hero's name.
    village: :class:`VillageType`
        The village type (home or builder base) where this hero belongs.
    is_home_base: :class:`bool`
        Whether this hero belongs to the home village.
    is_builder_base: :class:`bool`
        Whether this hero belongs to the builder base.
    max_level: :class:`int`
        The maximum level this hero can be upgraded to.
    id: :class:`int`
        The hero's unique identifier.
    info: :class:`str`
        Description of the hero.
    TID: :class:`TID`
        The hero's translation IDs for localization.
    production_building: :class:`ProductionBuildingType`
        The building type that produces this hero.
    production_building_level: :class:`int`
        The required level of the production building to unlock this hero.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this hero.
    is_flying: :class:`bool`
        Whether this hero is an air unit.
    is_air_targeting: :class:`bool`
        Whether this hero can target air units.
    is_ground_targeting: :class:`bool`
        Whether this hero can target ground units.
    movement_speed: :class:`int`
        The hero's movement speed.
    attack_speed: :class:`int`
        The hero's attack speed.
    attack_range: :class:`int`
        The hero's attack range.
    hitpoints: :class:`int`
        The hero's hitpoints.
    dps: :class:`int`
        The hero's damage per second.
    upgrade_time: :class:`TimeDelta`
        The time required to upgrade to the next level.
    upgrade_cost: :class:`int`
        The cost to upgrade to the next level.
    required_hero_tavern_level: Optional[:class:`int`]
        The hero tavern level required to upgrade to the next level.
    required_townhall: :class:`int`
        The townhall level required to upgrade to the next level.
    """

    __slots__ = (
        "name",
        "village",
        "is_home_base",
        "is_builder_base",
        "max_level",
        "equipment",
        "id",
        "info",
        "TID",
        "production_building",
        "production_building_level",
        "upgrade_resource",
        "is_flying",
        "is_air_targeting",
        "is_ground_targeting",
        "movement_speed",
        "attack_speed",
        "attack_range",
        "hitpoints",
        "dps",
        "upgrade_time",
        "upgrade_cost",
        "required_hero_tavern_level",
        "required_townhall",
        "_raw_data"
    )

    def __init__(self, data: dict, static_data: dict | None, level: int = 0, client = None):
        super().__init__(
            initial_level=data.get("level") or level,
            static_data=static_data
        )

        if client and client.raw_attribute and data:
            self._raw_data = data

        if data:
            self.name: str = data["name"]
            self.village = VillageType(value=data["village"])
            self.max_level: int = data["maxLevel"]
            self.equipment: list['Equipment'] = [
                Equipment(data=e, static_data=None) for e in data.get("equipment", [])
            ]

        if static_data:
            self.id: int = static_data["_id"]
            self.name: str = static_data["name"]
            self.info: str = static_data["info"]
            self.TID: TID = TID(data=static_data["TID"])

            self.production_building = ProductionBuildingType(value=str(static_data["production_building"]))
            self.production_building_level: int = static_data["production_building_level"] or 0
            self.upgrade_resource: Resource = Resource(value=static_data["upgrade_resource"])

            self.is_flying: bool = static_data["is_flying"]
            self.is_air_targeting: bool = static_data["is_air_targeting"]
            self.is_ground_targeting: bool = static_data["is_ground_targeting"]
            self.movement_speed: int = static_data["movement_speed"]
            self.attack_speed: int = static_data["attack_speed"]
            self.attack_range: int = static_data["attack_range"]

            self.village = VillageType(value=static_data["village"])

            self._load_level_data()

        self.is_home_base: bool = self.village == VillageType.home
        self.is_builder_base: bool = self.village == VillageType.builder_base

    def _load_level_data(self):
        if not self._static_data:
            return

        start_level = self._static_data["levels"][0]["level"]
        level_data = self._static_data["levels"][self._level - start_level]

        self.hitpoints: int = level_data["hitpoints"]
        self.dps: int = level_data["dps"]
        self.upgrade_time = TimeDelta(seconds=level_data["upgrade_time"])
        self.upgrade_cost: int = level_data["upgrade_cost"]
        self.required_hero_tavern_level: int | None = level_data["required_hero_tavern_level"]
        self.required_townhall: int = level_data["required_townhall"]


class Pet(LeveledUnit):
    """Represents a Pet object as returned by the API, optionally filled with game data.

    Attributes
    ----------
    name: :class:`str`
        The pet's name.
    village: :class:`VillageType`
        The village type where this pet belongs.
    max_level: :class:`int`
        The maximum level this pet can be upgraded to.
    id: :class:`int`
        The pet's unique identifier.
    info: :class:`str`
        Description of the pet.
    TID: :class:`TID`
        The pet's translation IDs for localization.
    production_building: :class:`ProductionBuildingType`
        The building type that produces this pet.
    production_building_level: :class:`int`
        The required level of the production building to unlock this pet.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this pet.
    is_flying: :class:`bool`
        Whether this pet is an air unit.
    is_air_targeting: :class:`bool`
        Whether this pet can target air units.
    is_ground_targeting: :class:`bool`
        Whether this pet can target ground units.
    movement_speed: :class:`int`
        The pet's movement speed.
    attack_speed: :class:`int`
        The pet's attack speed.
    attack_range: :class:`int`
        The pet's attack range.
    hitpoints: :class:`int`
        The pet's hitpoints.
    dps: :class:`int`
        The pet's damage per second.
    upgrade_time: :class:`TimeDelta`
        The time required to upgrade to the next level.
    upgrade_cost: :class:`int`
        The cost to upgrade to the next level.
    required_pet_house_level: Optional[:class:`int`]
        The pet house level required to upgrade to the next level.
    required_townhall: :class:`int`
        The townhall level required to upgrade to the next level.
    """

    __slots__ = (
        "name",
        "village",
        "max_level",
        "id",
        "info",
        "TID",
        "production_building",
        "production_building_level",
        "upgrade_resource",
        "is_flying",
        "is_air_targeting",
        "is_ground_targeting",
        "movement_speed",
        "attack_speed",
        "attack_range",
        "hitpoints",
        "dps",
        "upgrade_time",
        "upgrade_cost",
        "required_pet_house_level",
        "required_townhall",
        "_raw_data"
    )

    def __init__(self, data: dict, static_data: dict | None, level: int = 0, client = None):
        super().__init__(
            initial_level=data.get("level") or level,
            static_data=static_data
        )

        if client and client.raw_attribute and data:
            self._raw_data = data

        if data:
            self.name: str = data["name"]
            self.village = VillageType(value=data["village"])
            self.max_level: int = data["maxLevel"]

        if static_data:
            self.id: int = static_data["_id"]
            self.name: str = static_data["name"]
            self.info: str = static_data["info"]
            self.TID: TID = TID(data=static_data["TID"])

            self.production_building = ProductionBuildingType(value=static_data["production_building"])
            self.production_building_level: int = static_data["production_building_level"] or 0
            self.upgrade_resource: Resource = Resource(value=static_data["upgrade_resource"])

            self.is_flying: bool = static_data["is_flying"]
            self.is_air_targeting: bool = static_data["is_air_targeting"]
            self.is_ground_targeting: bool = static_data["is_ground_targeting"]
            self.movement_speed: int = static_data["movement_speed"]
            self.attack_speed: int = static_data["attack_speed"]
            self.attack_range: int = static_data["attack_range"]

            self.village = VillageType.home

            self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.hitpoints: int = level_data["hitpoints"]
        self.dps: int = level_data["dps"]
        self.upgrade_time = TimeDelta(seconds=level_data["upgrade_time"])
        self.upgrade_cost: int = level_data["upgrade_cost"]
        self.required_pet_house_level: int | None = level_data["required_pet_house_level"]
        self.required_townhall: int = level_data["required_townhall"]


class Equipment(LeveledUnit):
    """Represents an Equipment object as returned by the API, optionally filled with game data.

    Attributes
    ----------
    name: :class:`str`
        The equipment's name.
    village: :class:`VillageType`
        The village type where this equipment belongs.
    max_level: :class:`int`
        The maximum level this equipment can be upgraded to.
    id: :class:`int`
        The equipment's unique identifier.
    info: :class:`str`
        Description of the equipment.
    TID: :class:`TID`
        The equipment's translation IDs for localization.
    production_building: :class:`ProductionBuildingType`
        The building type that produces this equipment.
    production_building_level: :class:`int`
        The required level of the production building to unlock this equipment.
    rarity: :class:`EquipmentRarity`
        The rarity tier of this equipment.
    hero: :class:`str`
        The hero this equipment belongs to.
    hitpoints: :class:`int`
        The equipment's hitpoints bonus.
    dps: :class:`int`
        The equipment's damage per second bonus.
    heal_on_activation: :class:`int`
        The amount of healing provided when the equipment ability is activated.
    required_blacksmith_level: :class:`int`
        The blacksmith level required to upgrade to the next level.
    required_townhall: :class:`int`
        The townhall level required to upgrade to the next level.
    shiny_ore: :class:`int`
        The amount of shiny ore required to upgrade to the next level.
    glowy_ore: :class:`int`
        The amount of glowy ore required to upgrade to the next level.
    starry_ore: :class:`int`
        The amount of starry ore required to upgrade to the next level.
    abilities: List[:class:`dict`]
        The list of abilities this equipment provides.
    """

    __slots__ = (
        "name",
        "village",
        "max_level",
        "id",
        "info",
        "TID",
        "production_building",
        "production_building_level",
        "rarity",
        "hero",
        "hitpoints",
        "dps",
        "heal_on_activation",
        "required_blacksmith_level",
        "required_townhall",
        "shiny_ore",
        "glowy_ore",
        "starry_ore",
        "abilities",
        "_raw_data"
    )

    def __init__(self, data: dict, static_data: dict | None, level: int = 0, client = None):
        super().__init__(
            initial_level=data.get("level") or level,
            static_data=static_data
        )

        if client and client.raw_attribute and data:
            self._raw_data = data

        if data:
            self.name: str = data["name"]
            self.village = VillageType(value=data["village"])
            self.max_level: int = data["maxLevel"]

        if static_data:
            self.id: int = static_data["_id"]
            self.name: str = static_data["name"]
            self.info: str = static_data["info"]
            self.TID = TID(data=static_data["TID"])

            self.production_building = ProductionBuildingType(value=static_data["production_building"])
            self.production_building_level: int = static_data["production_building_level"] or 0

            self.rarity = EquipmentRarity(value=static_data["rarity"])
            self.hero: str = static_data["hero"]

            self.village = VillageType.home

            self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.hitpoints: int = level_data["hitpoints"]
        self.dps: int = level_data["dps"]
        self.heal_on_activation: int = level_data["heal_on_activation"]
        self.required_blacksmith_level: int = level_data["required_blacksmith_level"]
        self.required_townhall: int = level_data["required_townhall"]

        self.shiny_ore: int = level_data["upgrade_cost"]["shiny_ore"]
        self.glowy_ore: int = level_data["upgrade_cost"]["glowy_ore"]
        self.starry_ore: int = level_data["upgrade_cost"]["starry_ore"]

        self.abilities: list[dict] = level_data.get("abilities", [])



import unittest

from coc.wars import ClanWar
from coc.miscmodels import Timestamp

from tests.mockdata.mock_current_war import MOCK_CURRENT_WAR_IN_WAR


class TestWars(unittest.TestCase):
    def test_state(self):
        data = [
            {"state": "notInWar"},
            {"state": "preparation"},
            {"state": "inWar"},
            {"state": "warEnded"},
            {},
        ]
        for case in data:
            war = ClanWar(data=case, clan_tag="", client=None)
            try:
                state = case["state"]
                self.assertIsInstance(state, str)
            except KeyError:
                state = None
                self.assertIsNone(war.state)
            self.assertEqual(war.state, state)

    def test_team_size(self):
        data = [{"teamSize": 15}, {"teamSize": 30}, {"teamSize": 5}, {}]
        for case in data:
            war = ClanWar(data=case, clan_tag="", client=None)
            try:
                size = case["teamSize"]
                self.assertIsInstance(war.team_size, int)
            except KeyError:
                size = 0
                self.assertEqual(war.team_size, 0)
            self.assertEqual(war.team_size, size)

    def test_start_times(self):
        data = [
            {
                "preparationStartTime": "20200522T051229.000Z",
                "startTime": "20200523T043025.000Z",
                "endTime": "20200524T043025.000Z",
            },
        ]
        for case in data:
            war = ClanWar(data=case, clan_tag="", client=None)

            self.assertGreater(war.start_time, war.preparation_start_time)
            self.assertGreater(war.end_time, war.start_time)

            self.assertLess(war.preparation_start_time, war.start_time)
            self.assertLess(war.start_time, war.end_time)

            self.assertIsInstance(war.preparation_start_time, Timestamp)
            self.assertIsInstance(war.start_time, Timestamp)
            self.assertIsInstance(war.end_time, Timestamp)

            self.assertEqual(war.type, "random")


class TestWarClan(unittest.TestCase):
    def test_level(self):
        data = {"clan": {"clanLevel": 10}}
        war = ClanWar(data=data, client=None, clan_tag=None)
        self.assertIsInstance(war.clan.level, int)

    def test_destruction(self):
        data = {"clan": {"destructionPercentage": 84.88}}
        war = ClanWar(data=data, client=None, clan_tag=None)
        self.assertIsInstance(war.clan.destruction, float)

    def test_exp_earned(self):
        data = {"clan": {"clanLevel": 10}}
        war = ClanWar(data=data, client=None, clan_tag=None)
        self.assertIsInstance(war.clan.level, int)

    def test_clan_level(self):
        data = {"clan": {"clanLevel": 10}}
        war = ClanWar(data=data, client=None, clan_tag=None)
        self.assertIsInstance(war.clan.level, int)

    def test_clan_level(self):
        data = {"clan": {"clanLevel": 10}}
        war = ClanWar(data=data, client=None, clan_tag=None)
        self.assertIsInstance(war.clan.level, int)

    def test_clan_level(self):
        data = {"clan": {"clanLevel": 10}}
        war = ClanWar(data=data, client=None, clan_tag=None)
        self.assertIsInstance(war.clan.level, int)

    def test_clan_level(self):
        data = {"clan": {"clanLevel": 10}}
        war = ClanWar(data=data, client=None, clan_tag=None)
        self.assertIsInstance(war.clan.level, int)





import asyncio

from collections.abc import Iterable

from .errors import Maintenance, NotFound, Forbidden


class _AsyncIterator:
    """Base class for all async iterators."""

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            msg = await self._next()
        except StopAsyncIteration:
            raise StopAsyncIteration()
        else:
            return msg

    async def flatten(self):
        """
        |coro|

        Flattens the async iterator into a :class:`list` with all the elements.

        Returns
        --------
        :class:`list` - A list of every element in the async iterator.
        """
        ret = []
        while True:
            try:
                msg = await self._next()
            except StopAsyncIteration:
                return ret
            else:
                ret.append(msg)

    async def _next(self):
        return


class TaggedIterator(_AsyncIterator):
    """Implements filling of the queue and fetching results."""

    def __init__(self, client, tags: Iterable, cls, **kwargs):
        # pylint: disable=too-many-arguments
        self.client = client
        self.tags = tags

        self.cls = cls
        self.kwargs = kwargs

        self.queue = asyncio.Queue()
        self.queue_empty = True

        self.get_method = None  # set in subclass

    async def _run_method(self, tag: str):
        # pylint: disable=not-callable
        try:
            # I'm yet to find a way to only pass an arg/kwarg if it's not None, so lets just do this in interim
            if self.cls:
                return await self.get_method(tag, cls=self.cls, **self.kwargs)
            return await self.get_method(tag, **self.kwargs)
        except (NotFound, Forbidden, Maintenance):
            return None

    async def _fill_queue(self):
        tasks = [self.client.loop.create_task(self._run_method(n)) for n in self.tags]

        results = await asyncio.gather(*tasks)

        for result in results:
            if result:
                await self.queue.put(result)

    async def _next(self):
        """Retrieves the next item from the queue. If empty, fill the queue first."""
        if self.queue_empty:
            try:
                await self._fill_queue()
            except KeyError:
                await self.client.reset_keys()
                return await self._next()

            self.queue_empty = False

        try:
            return self.queue.get_nowait()
        except asyncio.QueueEmpty:
            raise StopAsyncIteration

class SeasonIterator(_AsyncIterator):
    """Iterator for paginated Legend League season rankings."""

    def __init__(self, client, league_id: int, season_id: str, cls, limit=100, **kwargs):
        self.client = client
        self.league_id = league_id
        self.season_id = season_id
        self.cls = cls
        self.limit = limit
        self.kwargs = kwargs

        self.buffer: list = []
        self.after: str | None = None
        self.exhausted = False

    async def _fetch_next_page(self):
        if self.exhausted:
            raise StopAsyncIteration

        params = {
            **self.kwargs,
            "limit": self.limit
        }
        if self.after:
            params["after"] = self.after

        data = await self.client.http.get_league_season_info(
            self.league_id, self.season_id, **params
        )

        self.buffer = [
            self.cls(data=item, client=self.client)
            for item in data.get("items", [])
        ]

        # Get pagination cursor
        try:
            self.after = data["paging"]["cursors"]["after"]
        except KeyError:
            self.exhausted = True
            self.after = None

        if not self.buffer:
            self.exhausted = True
            raise StopAsyncIteration

    async def _next(self):
        while not self.buffer:
            await self._fetch_next_page()

        return self.buffer.pop(0)


class ClanIterator(TaggedIterator):
    """Iterator for use with :meth:`~coc.Client.get_clans`"""

    def __init__(self, client, tags: Iterable, cls=None, **kwargs):
        # pylint: disable=too-many-arguments
        super().__init__(client, tags, cls, **kwargs)
        self.get_method = client.get_clan


class PlayerIterator(TaggedIterator):
    """Iterator for use with :meth:`~coc.Client.get_players`"""

    def __init__(self, client, tags: Iterable, cls=None, **kwargs):
        # pylint: disable=too-many-arguments
        super().__init__(client, tags, cls, **kwargs)
        self.get_method = self.get_player

        self.members = kwargs.pop("members", {})

    async def get_player(self, tag, cls=None, **kwargs):
        if cls:
            player = await self.client.get_player(tag, cls=cls, **kwargs)
        else:
            player = await self.client.get_player(tag, **kwargs)

        player._inject_clan_member(self.members.get(tag))
        return player


class ClanWarIterator(TaggedIterator):
    """Iterator for use with :meth:`~coc.Client.get_clan_wars`"""

    def __init__(self, client, tags: Iterable, cls=None, **kwargs):
        # pylint: disable=too-many-arguments
        super().__init__(client, tags, cls, **kwargs)
        self.get_method = client.get_clan_war


class LeagueWarIterator(TaggedIterator):
    """Iterator for use with :meth:`~coc.Client.get_league_wars`"""

    def __init__(self, client, tags: Iterable, clan_tag=None, cls=None, **kwargs):
        # pylint: disable=too-many-arguments
        super().__init__(client, tags, cls, clan_tag=clan_tag, **kwargs)
        self.get_method = client.get_league_war
        self.clan_tag = clan_tag

    async def _next(self):
        war = await super()._next()
        if war is None:
            return None
        elif self.clan_tag is None:
            return war
        elif war.clan_tag != self.clan_tag:
            return await self._next()
        else:
            return war

class CurrentWarIterator(TaggedIterator):
    """Iterator for use with :meth:`~coc.Client.get_current_wars`"""

    def __init__(self, client, tags: Iterable, cls=None, **kwargs):
        # pylint: disable=too-many-arguments
        super().__init__(client, tags, cls, **kwargs)
        self.get_method = client.get_current_war




import typing
from typing import List, TYPE_CHECKING

from . import BasePlayer
from .miscmodels import Badge, Timestamp, try_enum
from .utils import cached_property, correct_tag

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .war_members import ClanWarMember  # noqa


class RaidMember(BasePlayer):
    """Represents a Raid Member that the API returns.

    Attributes
    ----------
    tag: :class:`str`
        The player's tag
    name: :class:`str`
        The player's name
    attack_count: :class:`int`
        The number of attacks from this player
    attack_limit: :class:`int`
        The limit of attacks from this player
    bonus_attack_limit: :class:`int`
        The limit of bonus attacks from this player
    capital_resources_looted:  :class:`int`
        The amount of resources looted by this player
    raid_log_entry: :class:`RaidLogEntry`
        The raid log entry this member is in
    """

    __slots__ = ("tag",
                 "name",
                 "attack_count",
                 "attack_limit",
                 "bonus_attack_limit",
                 "capital_resources_looted",
                 "raid_log_entry",
                 "_cs_attacks",
                 "_client")

    def __init__(self, *, data, client, raid_log_entry):
        super().__init__(data=data, client=client)
        self._client = client
        self.raid_log_entry = raid_log_entry
        self._from_data(data)

    def __repr__(self):
        attrs = [
            ("tag", self.tag),
            ("raid_log_entry", repr(self.raid_log_entry)),
            ("attack_count", self.attack_count)
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return (isinstance(other, RaidMember)
                and self.tag == other.tag
                and self.raid_log_entry == other.raid_log_entry)

    def _from_data(self, data):
        data_get = data.get

        self.name: str = data_get("name")
        self.tag: str = data_get("tag")
        self.attack_count: int = data_get("attacks")
        self.attack_limit: int = data_get("attackLimit")
        self.bonus_attack_limit: int = data_get("bonusAttackLimit")
        self.capital_resources_looted: int = data_get("capitalResourcesLooted")

    @cached_property("_cs_attacks")
    def attacks(self):
        """List[:class:`RaidAttack`]: The member's attacks in this raid log entry.
        Can be empty due to missing parts in the API response.
        """
        return list(attack for attack_raid in self.raid_log_entry.attack_log
                    for district in attack_raid.districts for attack in district.attacks
                    if attack and attack.attacker_tag == self.tag)


class RaidAttack:
    """Represents a Raid attack

    Attributes
    ----------
    attacker_tag:
        :class:`str` - The attacker tag
    attacker_name:
        :class:`str` - The attacker name
    destruction:
        :class:`float`- The destruction achieved
    raid_log_entry:
        :class:`RaidLogEntry` - The raid log entry this attack belongs to
    raid_clan:
        :class:`RaidClan` - The raid clan this attack belongs to
    district:
        :class:`RaidDistrict` - The raid district this attack belongs to
    stars:
        :class:`int` - The raid attacks stars
    """

    __slots__ = ("raid_log_entry",
                 "raid_clan",
                 "district",
                 "raid_member",
                 "attacker_tag",
                 "attacker_name",
                 "destruction",
                 "stars",
                 "_client",
                 "_raw_data",
                 )

    def __repr__(self):
        attrs = [
            ("raid_log_entry", repr(self.raid_log_entry)),
            ("raid_clan", repr(self.raid_clan)),
            ("district", repr(self.district)),
            ("attacker_tag", self.attacker_tag),
            ("destruction", self.destruction),
            ("stars", self.stars),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return (isinstance(other, RaidAttack)
                and self.attacker_tag == other.attacker_tag
                and self.destruction == other.destruction
                and self.raid_clan == other.raid_clan
                and self.district == other.district
                and self.stars == other.stars)

    def __init__(self, data, client, raid_log_entry, raid_clan, district):
        self.raid_log_entry = raid_log_entry
        self.raid_clan = raid_clan
        self.district = district
        self._client = client
        self._raw_data = data if client and client.raw_attribute else None
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        self.attacker_tag = data["attacker"]["tag"]
        self.attacker_name = data["attacker"]["name"]
        self.destruction = data["destructionPercent"]
        self.stars = data["stars"]

    @property
    def attacker(self) -> "RaidMember":
        """:class:`RaidMember`: Returns the attacking player."""
        return self.raid_log_entry.get_member(self.attacker_tag)


class RaidDistrict:
    """Represents a Raid Clan Capital District.

    Attributes
    -----------
    id:
        :class:`int`: The district's unique ID as given by the API.
    name:
        :class:`str`: The district's name.
    hall_level:
        :class:`str`: The district's hall level.
    destruction:
        :class:`float`: The districts destruction percentage
    attack_count:
        :class:`int`: The districts attack count
    looted:
        :class:`int`: The districts total looted
    attacks:
        List[:class:`RaidAttack`]: The attacks on this district. Can be empty due to missing parts in the api response
    raid_log_entry:
        :class:`RaidLogEntry` - The raid log entry this district belongs to
    raid_clan:
        :class:`RaidClan` - The raid clan this district belongs to
    """

    def __eq__(self, other):
        return (isinstance(other, RaidDistrict)
                and self.id == other.id
                and self.raid_clan == other.raid_clan)

    __slots__ = ("id",
                 "name",
                 "hall_level",
                 "stars",
                 "destruction",
                 "attack_count",
                 "looted",
                 "attacks",
                 "raid_log_entry",
                 "raid_clan",
                 "_client",
                 "_raw_data")

    def __str__(self):
        return self.name

    def __repr__(self):
        attrs = [("id", self.id),
                 ("raid_log_entry", repr(self.raid_log_entry)),
                 ("raid_clan", repr(self.raid_clan)),
                 ("hall_level", self.hall_level),
                 ("stars", self.stars),
                 ("destruction", self.destruction)]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __init__(self, *, data, client, raid_log_entry, raid_clan):
        self.id: int = data.get("id")
        self.name: str = data.get("name")
        self.hall_level: int = data.get("districtHallLevel")
        self.stars: int = data.get("stars")
        self.destruction: float = data.get("destructionPercent")
        self.attack_count: int = data.get("attackCount")
        self.looted: int = data.get("totalLooted")
        self.raid_log_entry = raid_log_entry  # type: RaidLogEntry
        self.raid_clan = raid_clan  # type: RaidClan
        self._raw_data = data if client and client.raw_attribute else None
        self._client = client
        if data.get("attacks", None):
            self.attacks: List[RaidAttack] = [RaidAttack(data=adata, client=client,
                                                         raid_log_entry=self.raid_log_entry,
                                                         raid_clan=self.raid_clan, district=self)
                                              for adata in data.get("attacks")]
        else:
            self.attacks = []
        if self.destruction != 0 and self.stars == 0:  # attempt to fix an api bug responding with the wrong star count
            self.stars = max([self.stars] + [a.stars for a in self.attacks if a])


class RaidClan:
    """Represents the clan object returned by clan raid seasons.

        Attributes
        ----------
        tag: :class:`str`
            The clan's tag
        name: :class:`str`
            The clan's name
        badge: :class:`Badge`
            The clan's badge
        level: :class:`int`
            The clan's level.
        attack_count: :class:`int`
            The number of attacks in the raid.
        district_count: :class:`int`
            The number of districts in the raid.
        destroyed_district_count: :class:`int`
            The number of destroyed districts in the raid.
        index:
            :class:`int` - The index/order of the raid clan in the raid weekend
        raid_log_entry:
            :class:`RaidLogEntry` - The raid log entry this attack belongs to
        """

    __slots__ = (
        "tag",
        "name",
        "badge",
        "level",
        "index",
        "attack_count",
        "district_count",
        "destroyed_district_count",
        "raid_log_entry",
        "_districts",
        "_attacks",
        "_client",
        "_response_retry",
        "_cs_raid_attacks",
        "_cs_raid_districts",
        "_cs_looted",
        "_iter_raid_districts",
        "_raw_data"
    )

    def __init__(self, *, data, client, raid_log_entry, index=0, **_):
        self._client = client
        self._raw_data = data if client and client.raw_attribute else None
        self._response_retry = data.get("_response_retry")
        self.tag = data.get("attacker", data.get("defender")).get("tag")
        self.name = data.get("attacker", data.get("defender")).get("name")
        self.badge = try_enum(Badge, data=data.get("attacker", data.get("defender")).get("badgeUrls"),
                              client=self._client)
        self.level = data.get("attacker", data.get("defender")).get("level")
        self.raid_log_entry = raid_log_entry
        self.index = index
        self._from_data(data)

    def __eq__(self, other):
        return (isinstance(other, RaidClan)
                and self.tag == other.tag
                and self.raid_log_entry.start_time == other.raid_log_entry.start_time
                and self.index == other.index)

    def __repr__(self):
        attrs = [
            ("tag", self.tag),
            ("name", self.name),
            ("raid_log_entry", repr(self.raid_log_entry))
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def _from_data(self, data):
        data_get = data.get

        self.attack_count: int = data_get("attackCount")
        self.district_count: int = data_get("districtCount")
        self.destroyed_district_count: int = data_get("districtsDestroyed")

        if data_get("districts"):
            self._iter_raid_districts = (RaidDistrict(data=data, client=self._client,
                                                      raid_log_entry=self.raid_log_entry, raid_clan=self) for
                                         data in data_get("districts"))
        else:
            self._iter_raid_districts = ()

    @cached_property("_cs_raid_districts")
    def districts(self) -> typing.List[RaidDistrict]:
        """List[:class:`RaidDistrict`]: A :class:`List` of :class:`RaidDistrict` that the clan
        attacked."""
        return list(self._iter_raid_districts)

    @cached_property("_cs_raid_attacks")
    def attacks(self) -> typing.List[RaidAttack]:
        """List[:class:`RaidAttack`]: Returns all attacks in the raid against this clan."""
        return list(attack for district in self.districts for attack in district.attacks)

    @cached_property("_cs_looted")
    def looted(self) -> int:
        """:class:`int`: The amount of loot the raid clan got on all districts."""
        loot = 0
        for district in self.districts:
            loot += district.looted
        return loot

    @property
    def is_finished(self) -> bool:
        return self.destroyed_district_count == self.district_count


class RaidLogEntry:
    """Represents a Clash of Clans Raid Log Entry

    Attributes
    ----------
    state:
        :class:`str`: The state of the raid log entry. Currently, the states ``ongoing`` and ``ended`` are known.
    start_time:
        :class:`Timestamp`: The :class:`Timestamp` that the raid started at.
    end_time:
        :class:`Timestamp`: The :class:`Timestamp` that the raid ended at.
    total_loot:
        :class:`int`: The amount of total loot
    completed_raid_count:
        :class:`int`: The number of completed raids
    attack_count:
        :class:`int`: The total number of attacks
    destroyed_district_count:
        :class:`int`: The number of destroyed enemy districts
    offensive_reward:
        :class:`int`: The amount of offensive reward
    defensive_reward:
        :class:`int`: The amount of defensive reward
    """

    __slots__ = ("clan_tag",
                 "state",
                 "start_time",
                 "end_time",
                 "total_loot",
                 "completed_raid_count",
                 "attack_count",
                 "destroyed_district_count",
                 "offensive_reward",
                 "defensive_reward",
                 "_raw_data",
                 "_cs_attack_log",
                 "_cs_defense_log",
                 "_cs_members",
                 "_cs_total_defensive_loot",
                 "_cs_defense_attack_count",
                 "_cs_defensive_destroyed_district_count",
                 "_iter_members",
                 "_iter_attack_log",
                 "_iter_defense_log",
                 "_members",
                 "_attack_log",
                 "_defense_log",
                 "_client",
                 "_response_retry"
                 )

    def __init__(self, *, data, client, **kwargs):
        self._client = client
        self.clan_tag = kwargs['clan_tag'] if "clan_tag" in kwargs else ""
        self._response_retry = kwargs['response_retry'] if "response_retry" in kwargs else 0
        self._raw_data = data if client and client.raw_attribute else None
        self._from_data(data)
        self._members = {}
        self._attack_log = []
        self._defense_log = []

    def __repr__(self):
        attrs = [
            ("state", self.state),
            ("start_time", self.start_time),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return (isinstance(other, RaidLogEntry)
                and self.start_time == other.start_time
                and self.clan_tag == other.clan_tag)

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.state: str = data_get("state")
        self.start_time = try_enum(Timestamp, data=data_get("startTime"))
        self.end_time = try_enum(Timestamp, data=data_get("endTime"))
        self.total_loot: int = data_get("capitalTotalLoot")
        self.completed_raid_count: int = data_get("raidsCompleted")
        self.attack_count: int = data_get("totalAttacks")
        self.destroyed_district_count: int = data_get("enemyDistrictsDestroyed")
        self.offensive_reward: int = data_get("offensiveReward")
        self.defensive_reward: int = data_get("defensiveReward")

        self._iter_attack_log = (RaidClan(data=adata, raid_log_entry=self, client=self._client, index=c)
                                 for c, adata in enumerate(data_get("attackLog", [])))

        self._iter_defense_log = (RaidClan(data=adata, raid_log_entry=self, client=self._client, index=c)
                                  for c, adata in enumerate(data_get("defenseLog", [])))

        self._iter_members = (RaidMember(data=adata, raid_log_entry=self, client=self._client)
                              for adata in data_get("members", []))

    @cached_property("_cs_members")
    def members(self) -> typing.List[RaidMember]:
        """List[:class:`RaidMember`]: A list of members that are in the raid.
        """
        dict_members = self._members = {m.tag: m for m in self._iter_members}
        return list(dict_members.values())

    @cached_property("_cs_attack_log")
    def attack_log(self) -> typing.List[RaidClan]:
        """List[:class:`RaidClan`]: A list of raid clans that are attacked in the raid season.
        """
        list_attack_log = self._attack_log = [m for m in self._iter_attack_log]
        return list_attack_log

    @cached_property("_cs_defense_log")
    def defense_log(self) -> typing.List[RaidClan]:
        """List[:class:`RaidClan`]: A list of raid clans which represents all the defensive raids of a season.
        """
        list_defense_log = self._defense_log = [m for m in self._iter_defense_log]
        return list_defense_log

    @cached_property("_cs_total_defensive_loot")
    def total_defensive_loot(self) -> int:
        """:class:`int`: The total amount of loot taken by all opponents of the raid weekend."""
        loot = 0
        for clan in self.defense_log:
            loot += clan.looted
        return loot

    @cached_property("_cs_defense_attack_count")
    def defense_attack_count(self) -> int:
        """:class:`int`: The total amount of opponent attacks in the raid weekend."""
        count = 0
        for clan in self.defense_log:
            count += clan.attack_count
        return count

    @cached_property("_cs_defensive_destroyed_district_count")
    def defensive_destroyed_district_count(self) -> int:
        """:class:`int`: The total amount of districts destroyed by opponents."""
        count = 0
        for clan in self.defense_log:
            count += clan.destroyed_district_count
        return count

    def get_member(self, tag: str) -> typing.Optional[RaidMember]:
        """Get a member of the clan for the given tag, or ``None`` if not found.

        Returns
        --------
        The clan member who matches the tag.: Optional[:class:`RaidMember`]"""
        tag = correct_tag(tag)
        if not self._members:
            _ = self.members

        try:
            return self._members[tag]
        except KeyError:
            return None




from .abc import LeveledUnit
from .enums import Resource, VillageType, ProductionBuildingType
from .miscmodels import TimeDelta, TID


class Troop(LeveledUnit):
    """Represents a Troop object as returned by the API, optionally filled with game data.

    Attributes
    ----------
    name: :class:`str`
        The troop's name.
    village: :class:`VillageType`
        The village type (home or builder base) where this troop belongs.
    is_home_base: :class:`bool`
        Whether this troop belongs to the home village.
    is_builder_base: :class:`bool`
        Whether this troop belongs to the builder base.
    max_level: :class:`int`
        The maximum level this troop can be upgraded to.
    is_active: :class:`bool`
        Whether this super troop is currently active/boosted.
    id: :class:`int`
        The troop's unique identifier.
    info: :class:`str`
        Description of the troop.
    TID: :class:`TID`
        The troop's translation IDs for localization.
    production_building: :class:`ProductionBuildingType`
        The building type that produces this troop.
    production_building_level: :class:`int`
        The required level of the production building to unlock this troop.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this troop.
    is_flying: :class:`bool`
        Whether this troop is an air unit.
    is_air_targeting: :class:`bool`
        Whether this troop can target air units.
    is_ground_targeting: :class:`bool`
        Whether this troop can target ground units.
    movement_speed: :class:`int`
        The troop's movement speed.
    attack_speed: :class:`int`
        The troop's attack speed.
    attack_range: :class:`int`
        The troop's attack range.
    housing_space: :class:`int`
        The housing space this troop occupies.
    is_super_troop: :class:`bool`
        Whether this is a super troop.
    is_seasonal: :class:`bool`
        Whether this is a seasonal/temporary troop.
    is_siege_machine: :class:`bool`
        Whether this is a siege machine.
    base_troop_id: :class:`int`
        The ID of the base troop (only for super troops).
    base_troop_minimum_level: :class:`int`
        The minimum level of the base troop required to boost this super troop.
    hitpoints: :class:`int`
        The troop's hitpoints.
    dps: :class:`int`
        The troop's damage per second.
    upgrade_time: :class:`TimeDelta`
        The time required to upgrade to the next level.
    upgrade_cost: :class:`int`
        The cost to upgrade to the next level.
    required_lab_level: Optional[:class:`int`]
        The laboratory level required to upgrade to the next level.
    required_townhall: :class:`int`
        The townhall level required to upgrade to the next level.
    """

    __slots__ = (
        "name",
        "village",
        "is_home_base",
        "is_builder_base",
        "max_level",
        "is_active",
        "id",
        "info",
        "TID",
        "production_building",
        "production_building_level",
        "upgrade_resource",
        "is_flying",
        "is_air_targeting",
        "is_ground_targeting",
        "movement_speed",
        "attack_speed",
        "attack_range",
        "housing_space",
        "is_super_troop",
        "is_seasonal",
        "is_siege_machine",
        "base_troop_id",
        "base_troop_minimum_level",
        "hitpoints",
        "dps",
        "upgrade_time",
        "upgrade_cost",
        "required_lab_level",
        "required_townhall",
        "_raw_data",
        "__data"
    )

    def __init__(self, data: dict, static_data: dict | None, level: int = 0, client = None):
        super().__init__(
            initial_level=data.get("level") or level,
            static_data=static_data
        )

        self.__data = data
        if client and client.raw_attribute and data:
            self._raw_data = data

        if data:
            self.name: str = data["name"]
            self.village = VillageType(value=data["village"])
            self.max_level: int = data["maxLevel"]
            self.is_active: bool = data.get("superTroopIsActive", False)

        if static_data:
            self.id: int = static_data["_id"]
            self.name: str = static_data["name"]
            self.info: str = static_data["info"]
            self.TID: TID = TID(data=static_data["TID"])

            self.production_building = ProductionBuildingType(value=static_data["production_building"])
            self.production_building_level: int = static_data["production_building_level"]
            self.upgrade_resource = Resource(value=static_data["upgrade_resource"])

            self.is_flying: bool = static_data["is_flying"]
            self.is_air_targeting: bool = static_data["is_air_targeting"]
            self.is_ground_targeting: bool = static_data["is_ground_targeting"]

            self.movement_speed: int = static_data["movement_speed"]
            self.attack_speed: int = static_data["attack_speed"]
            self.attack_range: int = static_data["attack_range"]
            self.housing_space: int = static_data["housing_space"]

            self.village = VillageType(value=static_data["village"])

            self.is_super_troop: bool = "super_troop" in static_data
            self.is_seasonal: bool = static_data.get("is_seasonal", False)
            self.is_siege_machine: bool = self.production_building == ProductionBuildingType.workshop

            if self.is_super_troop:
                self.base_troop_id: int = static_data["super_troop"]["original_id"]
                self.base_troop_minimum_level: int = static_data["super_troop"]["original_min_level"]

        self.is_home_base: bool = self.village == VillageType.home
        self.is_builder_base: bool = self.village == VillageType.builder_base

        self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        start_level = self._static_data["levels"][0]["level"]
        # hacky way to deal with lv 1 builder base troops showing as lv 1 instead of correct level
        if self.__data and self._level == 1 and self.is_builder_base:
            self._level = self._static_data["levels"][0]["level"] + self._level - 1

        level_data = self._static_data["levels"][self._level - start_level]

        self.hitpoints: int = level_data["hitpoints"]
        self.dps: int = level_data["dps"]
        self.upgrade_time = TimeDelta(seconds=level_data["upgrade_time"])
        self.upgrade_cost: int = level_data["upgrade_cost"]

        #is None for seasonal troops
        self.required_lab_level: int | None = level_data["required_lab_level"]
        self.required_townhall: int = level_data["required_townhall"]



from .characters import Guardian, Helper
from .buildings import Trap, Building, SeasonalDefense, SeasonalDefenseModule
from .cosmetics import Decoration, Obstacle, Skin, Scenery, ClanCapitalHousePart
from. hero import Hero, Pet, Equipment
from .troop import Troop
from .spell import Spell
from .abc import LeveledUnit
from .miscmodels import TimeDelta
from .enums import BuildingType

import re

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from client import Client

TROOP_BASE_ID = 4000000
SPELL_BASE_ID = 26000000

HERO_BASE_ID = 28000000
PET_BASE_ID = 73000000
EQUIPMENT_BASE_ID = 90000000


class Upgrade:
    """Represents an ongoing upgrade.

    Attributes
    ----------
    is_goblin: :class:`bool`
        Whether this is a goblin builder upgrade.
    target: :class:`LeveledUnit`
        The target item being upgraded.
    timer: :class:`TimeDelta`
        The time remaining for the upgrade.
    helper_timer: Optional[:class:`TimeDelta`]
        The time remaining for the helper boost.
    recurrent_helper: :class:`bool`
        Whether the helper is recurrent.
    """

    __slots__ = (
        "is_goblin",
        "target",
        "timer",
        "helper_timer",
        "recurrent_helper",
    )

    def __init__(self,
                 is_goblin: bool,
                 target: LeveledUnit,
                 timer: TimeDelta,
                 helper_timer: TimeDelta | None,
                 recurrent_helper: bool = False
                 ):
        self.is_goblin: bool = is_goblin
        self.target = target
        self.timer = timer
        self.helper_timer = helper_timer
        self.recurrent_helper = recurrent_helper

    def __repr__(self):
        attrs = [
            ("target", self.target.name if hasattr(self.target, "name") else "Unknown"),
            ("timer", self.timer),
            ("helper_timer", self.helper_timer),
            ("is_goblin", self.is_goblin),
            ("recurrent_helper", self.recurrent_helper),
        ]
        return "<%s %s>" % (
            self.__class__.__name__, " ".join("%s=%r" % t for t in attrs))


class Boosts:
    """Represents active boosts and cooldowns.

    Attributes
    ----------
    builder_boost: :class:`TimeDelta`
        The time remaining for the builder potion boost.
    lab_boost: :class:`TimeDelta`
        The time remaining for the research potion boost.
    clocktower_boost: :class:`TimeDelta`
        The time remaining for the clock tower boost.
    clocktower_cooldown: :class:`TimeDelta`
        The cooldown time for the clock tower.
    builder_consumable: :class:`TimeDelta`
        The time remaining for the builder bite snack.
    lab_consumable: :class:`TimeDelta`
        The time remaining for the study soup snack.
    helper_cooldown: :class:`TimeDelta`
        The global cooldown timer for helpers.
    """

    __slots__ = (
        "builder_boost",
        "lab_boost",
        "clocktower_boost",
        "clocktower_cooldown",
        "builder_consumable",
        "lab_consumable",
        "helper_cooldown",
    )

    def __init__(self, data: dict = {}):
        self.builder_boost: TimeDelta = TimeDelta(seconds=data.get("builder_boost", 0))
        self.lab_boost: TimeDelta = TimeDelta(seconds=data.get("lab_boost", 0))
        self.clocktower_boost: TimeDelta = TimeDelta(seconds=data.get("clocktower_boost", 0))
        self.clocktower_cooldown: TimeDelta = TimeDelta(seconds=data.get("clocktower_cooldown", 0))
        self.builder_consumable: TimeDelta = TimeDelta(seconds=data.get("builder_consumable", 0))
        self.lab_consumable: TimeDelta = TimeDelta(seconds=data.get("lab_consumable", 0))
        self.helper_cooldown: TimeDelta = TimeDelta(seconds=data.get("helper_cooldown", 0))

    def __repr__(self):
        attrs = [
            ("builder_boost", self.builder_boost),
            ("lab_boost", self.lab_boost),
            ("clocktower_boost", self.clocktower_boost),
            ("clocktower_cooldown", self.clocktower_cooldown),
            ("builder_consumable", self.builder_consumable),
            ("lab_consumable", self.lab_consumable),
            ("helper_cooldown", self.helper_cooldown),
        ]
        return "<%s %s>" % (
            self.__class__.__name__, " ".join("%s=%r" % t for t in attrs))


class StaticData:
    """Represents static game data loaded from game files.
    
    This class loads and organizes all static game data from the game's data files.
    It provides access to all available game items at all levels, useful for looking
    up item information, max levels, and statistics.
    
    Attributes
    ----------
    buildings: List[:class:`Building`]
        List of all available buildings.
    capital_house_parts: List[:class:`ClanCapitalHousePart`]
        List of all available clan capital house parts.
    decorations: List[:class:`Decoration`]
        List of all available decorations.
    equipment: List[:class:`Equipment`]
        List of all available equipment.
    guardians: List[:class:`Guardian`]
        List of all available guardians.
    helpers: List[:class:`Helper`]
        List of all available helpers.
    heroes: List[:class:`Hero`]
        List of all available heroes.
    obstacles: List[:class:`Obstacle`]
        List of all available obstacles.
    pets: List[:class:`Pet`]
        List of all available pets.
    sceneries: List[:class:`Scenery`]
        List of all available sceneries.
    skins: List[:class:`Skin`]
        List of all available skins.
    spells: List[:class:`Spell`]
        List of all available spells.
    traps: List[:class:`Trap`]
        List of all available traps.
    troops: List[:class:`Troop`]
        List of all available troops & siege machines.
    """
    
    __slots__ = (
        "_data",
        "helpers",
        "guardians",
        "buildings",
        "traps",
        "decorations",
        "obstacles",
        "troops",
        "siege_machines",
        "heroes",
        "spells",
        "pets",
        "equipment",
        "capital_house_parts",
        "skins",
        "sceneries",
    )
    
    def __init__(self, data: dict):
        self._data = data

        self.buildings: list[Building] = []
        self.capital_house_parts: list[ClanCapitalHousePart] = []
        self.decorations: list[Decoration] = []
        self.equipment: list[Equipment] = []
        self.guardians: list[Guardian] = []
        self.helpers: list[Helper] = []
        self.heroes: list[Hero] = []
        self.obstacles: list[Obstacle] = []
        self.pets: list[Pet] = []
        self.sceneries: list[Scenery] = []
        self.skins: list[Skin] = []
        self.spells: list[Spell] = []
        self.traps: list[Trap] = []
        self.troops: list[Troop] = []

        self._load_data()

    def _load_data(self):

        section_class_mapping = {
            "helpers": Helper,
            "buildings": Building,
            "traps": Trap,
            "troops": Troop,
            "guardians": Guardian,
            "spells": Spell,
            "heroes": Hero,
            "pets": Pet,
            "equipment": Equipment,
            "decorations": Decoration,
            "obstacles": Obstacle,
            "sceneries": Scenery,
            "skins": Skin,
            "capital_house_parts": ClanCapitalHousePart,
        }
        for section, items in self._data.items():
            cls = section_class_mapping.get(section)
            if cls is None:
                continue

            for item in items:
                data = item
                if section in ["troops", "spells", "heroes", "pets", "equipment"]:
                    static_data = item
                    data = {}
                    self.__getattribute__(section).append(cls(data=data, static_data=static_data))
                else:
                    self.__getattribute__(section).append(cls(data=data))


class AccountData:
    """Represents player account data parsed from game files.
    
    Parses raw account data from game files and creates game objects representing
    the player's village state, including buildings, troops, heroes, ongoing upgrades,
    and active boosts. This provides a complete snapshot of a player's account.

    Attributes
    ----------
    townhall_level: :class:`int`
        The player's current townhall level.
    buildings: List[Tuple[:class:`Building`, :class:`int`]]
        Player's buildings as tuples of (building, quantity). Buildings can have multiple
        instances (e.g., multiple cannons).
    capital_house_parts: List[:class:`ClanCapitalHousePart`]
        Player's unlocked clan capital house parts.
    decorations: List[Tuple[:class:`Decoration`, :class:`int`]]
        Player's decorations as tuples of (decoration, quantity).
    equipment: List[:class:`Equipment`]
        Player's equipment with their current levels.
    guardians: List[:class:`Guardian`]
        Player's guardians with their current levels.
    helpers: List[:class:`Helper`]
        Player's helpers with their current levels.
    heroes: List[:class:`Hero`]
        Player's heroes with their current levels.
    obstacles: List[Tuple[:class:`Obstacle`, :class:`int`]]
        Player's obstacles as tuples of (obstacle, quantity).
    pets: List[:class:`Pet`]
        Player's pets with their current levels.
    sceneries: List[:class:`Scenery`]
        Player's unlocked base sceneries.
    skins: List[:class:`Skin`]
        Player's unlocked hero skins.
    spells: List[:class:`Spell`]
        Player's unlocked spells with their current levels.
    traps: List[Tuple[:class:`Trap`, :class:`int`]]
        Player's traps as tuples of (trap, quantity).
    troops: List[:class:`Troop`]
        Player's unlocked troops with their current levels.
    upgrades: List[:class:`Upgrade`]
        Currently ongoing upgrades (buildings, troops, spells, heroes, etc.).
    boosts: :class:`Boosts`
        Active boosts and cooldowns (builder boost, lab boost, clock tower, etc.).
    """

    __slots__ = (
        "_client",
        "townhall_level",
        "helpers",
        "guardians",
        "buildings",
        "traps",
        "decorations",
        "obstacles",
        "troops",
        "heroes",
        "spells",
        "pets",
        "equipment",
        "capital_house_parts",
        "skins",
        "sceneries",
        "upgrades",
        "boosts",
    )

    def __init__(self, data: dict, client: 'Client'):
        self._client = client
        self.townhall_level: int = 0

        self.buildings: list[tuple[Building, int]] = []
        self.capital_house_parts: list[ClanCapitalHousePart] = []
        self.decorations: list[tuple[Decoration, int]] = []
        self.equipment: list[Equipment] = []
        self.guardians: list[Guardian] = []
        self.helpers: list[Helper] = []
        self.heroes: list[Hero] = []
        self.obstacles: list[tuple[Obstacle, int]] = []
        self.pets: list[Pet] = []
        self.sceneries: list[Scenery] = []
        self.skins: list[Skin] = []
        self.spells: list[Spell] = []
        self.traps: list[tuple[Trap, int]] = []
        self.troops: list[Troop] = []
        self.upgrades: list[Upgrade] = []

        self.boosts: Boosts = Boosts()

        self._load_data(account_data=data)

    def __repr__(self):
        attrs = []
        for attr_name in dir(self):
            if not attr_name.startswith('_') and not callable(getattr(self, attr_name)):
                attr_value = getattr(self, attr_name)
                if isinstance(attr_value, list):
                    attrs.append((attr_name, [repr(item) for item in attr_value]))
                else:
                    attrs.append((attr_name, repr(attr_value)))

        lines = [f"<{self.__class__.__name__}"]
        for name, value in attrs:
            lines.append(f"  {name}={value}")
        lines.append(">")
        return "\n".join(lines)

    def get_static_data_item(self, item_id: int | tuple[int, str]):
        item_data = self._client._static_data.get(item_id)
        return item_data

    def add_upgrade(self, item: dict, target: LeveledUnit) -> Upgrade | None:
        if "timer" in item:
            helper_timer = None
            if "helper_timer" in item:
                helper_timer = TimeDelta(seconds=item["helper_timer"])
            self.upgrades.append(
                Upgrade(
                    is_goblin=item.get("extra", False),
                    target=target,
                    timer=TimeDelta(seconds=item["timer"]),
                    helper_timer=helper_timer,
                    recurrent_helper=item.get("recurrent_helper", False)
                )
            )

    def _load_data(self, account_data: dict):
        self.boosts = Boosts(data=account_data.get("boosts", {}))

        # section is "guardians", "buildings", etc,
        for section, items in account_data.items():  # type: str, list[dict]

            if not isinstance(items, list) or "2" in section:
                continue
            # this adds builder base troops, heroes etc to the home village counterpart data/lists
            items.extend(account_data.get(f"{section}2", []))

            if section == "helpers":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    helper = Helper(
                        level=item.get("lvl", 0), data=item_data
                    )
                    if "helper_cooldown" in item:
                        self.boosts.helper_cooldown = TimeDelta(seconds=item["helper_cooldown"])
                    self.helpers.append(helper)

            elif section == "guardians":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    guardian = Guardian(
                        level=item.get("lvl"), data=item_data
                    )
                    self.add_upgrade(item, guardian)
                    self.guardians.append(guardian)

            elif section == "buildings":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    seasonal_defenses = []
                    if "types" in item:
                        crafting_station: list[dict] = item["types"]
                        for seasonal_defense in crafting_station:
                            seasonal_def_data = next((
                                item for item in item_data["seasonal_defenses"]
                                if item["_id"] == seasonal_defense["data"]
                            ))
                            if seasonal_def_data is None:
                                continue
                            modules = []
                            for module in seasonal_defense["modules"]:
                                module_data = next((
                                    item for item in seasonal_def_data["modules"] if item["_id"] == module["data"]
                                ))
                                if module_data is None:
                                    continue
                                modules.append(SeasonalDefenseModule(level=module["lvl"], data=module_data))
                                self.add_upgrade(item, module)
                            seasonal_defenses.append(SeasonalDefense(data=seasonal_def_data, modules=modules))

                    building = Building(
                        level=item.get("lvl", 0),
                        data=item_data,
                        weapon_level=item.get("weapon", 0),
                        supercharge_level=item.get("supercharge", 0),
                        seasonal_defenses=seasonal_defenses
                    )
                    if building.type == BuildingType.town_hall:
                        self.townhall_level = building.level
                    self.add_upgrade(item, building)
                    self.buildings.append((building, item.get("cnt", 1)))

            elif section == "traps":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    trap = Trap(
                        level=item.get("lvl"), data=item_data
                    )
                    self.add_upgrade(item, trap)
                    self.traps.append((trap, item.get("cnt", 1)))

            elif section == "decos":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    deco = Decoration(data=item_data)
                    self.decorations.append((deco, item.get("cnt", 1)))

            elif section == "obstacles":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    obstacle = Obstacle(data=item_data)
                    self.obstacles.append((obstacle, item.get("cnt", 1)))

            elif section == "units" or section == "siege_machines":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    troop = Troop(data={}, static_data=item_data, level=item["lvl"])
                    self.add_upgrade(item, troop)
                    self.troops.append(troop)

            elif section == "spells":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    spell = Spell(data={}, static_data=item_data, level=item["lvl"])
                    self.add_upgrade(item, spell)
                    self.spells.append(spell)

            elif section == "heroes":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    hero = Hero(data={}, static_data=item_data, level=item["lvl"])
                    self.add_upgrade(item, hero)
                    self.heroes.append(hero)

            elif section == "pets":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    pet = Pet(data={}, static_data=item_data, level=item["lvl"])
                    self.add_upgrade(item, pet)
                    self.pets.append(pet)

            elif section == "equipment":
                for item in items:
                    item_id = item["data"]
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    equipment = Equipment(data={}, static_data=item_data, level=item["lvl"])
                    self.equipment.append(equipment)

            elif section == "skins":
                for item_id in items:
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    skin = Skin(data=item_data)
                    self.skins.append(skin)

            elif section == "sceneries":
                for item_id in items:
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    scenery = Scenery(data=item_data)
                    self.sceneries.append(scenery)

            elif section == "house_parts":
                for item_id in items:
                    item_data = self.get_static_data_item(item_id=item_id)
                    if item_data is None:
                        continue
                    house_part = ClanCapitalHousePart(data=item_data)
                    self.capital_house_parts.append(house_part)


class HeroLoadout:
    """Represents a hero loadout from an army link.

    Attributes
    ----------
    hero: :class:`Hero`
        The hero in this loadout.
    pet: Optional[:class:`Pet`]
        The pet assigned to this hero.
    equipment: List[:class:`Equipment`]
        The list of equipment assigned to this hero.
    """

    __slots__ = (
        "_lookup",
        "hero",
        "pet",
        "equipment",
    )

    def __init__(self, loadout: tuple, lookup):

        self._lookup = lookup

        self.hero: Hero = Hero(
            data={}, level=1,
            static_data=self._lookup.get(loadout[0], self._lookup.get(HERO_BASE_ID))
        )
        self.pet = None
        if loadout[1]:
            self.pet: Pet | None = Pet(
                data={}, level=1,
                static_data=self._lookup.get(loadout[1], self._lookup.get(PET_BASE_ID))
            )
        self.equipment: list[Equipment] = []
        for e in loadout[2:]:
            if e:
                self.equipment.append(
                    Equipment(
                        data={}, level=1,
                        static_data=self._lookup.get(e, self._lookup.get(EQUIPMENT_BASE_ID)))
                    )


class ArmyRecipe:
    """Represents an army composition parsed from an army link.

    Attributes
    ----------
    link: :class:`str`
        The army link string.
    heroes_loadout: List[:class:`HeroLoadout`]
        The list of hero loadouts.
    troops: List[Tuple[:class:`Troop`, :class:`int`]]
        The list of troops with their quantities.
    spells: List[Tuple[:class:`Spell`, :class:`int`]]
        The list of spells with their quantities.
    clan_castle_troops: List[Tuple[:class:`Troop`, :class:`int`]]
        The list of clan castle troops with their quantities.
    clan_castle_spells: List[Tuple[:class:`Spell`, :class:`int`]]
        The list of clan castle spells with their quantities.
    """

    __slots__ = (
        "link",
        "_lookup",
        "heroes_loadout",
        "troops",
        "spells",
        "clan_castle_troops",
        "clan_castle_spells",
    )

    def __init__(self, static_data: dict, link: str):
        self.link = link

        self._lookup = static_data

        self.heroes_loadout: list[HeroLoadout] = []
        self.troops: list[tuple[Troop, int]] = []
        self.spells: list[tuple[Spell, int]] = []
        self.clan_castle_troops: list[tuple[Troop, int]] = []
        self.clan_castle_spells: list[tuple[Spell, int]] = []

        self._parse()

    def _parse_items(self, match_str: str, base_id: int, item_class, target_list: list):
        """Helper to parse troops or spells from army link."""
        for split in (item.split('x') for item in match_str.split('-')):
            item_id = int(split[1]) + base_id
            quantity = int(split[0])
            static_data = self._lookup.get(item_id, self._lookup.get(base_id))
            item = item_class(data={}, static_data=static_data, level=1)
            target_list.append((item, quantity))

    def _parse(self):

        ARMY_LINK_SEPARATOR = re.compile(
            r"h(?P<heroes>[^idus]+)"
            r"|i(?P<castle_troops>[\d+x-]+)"
            r"|d(?P<castle_spells>[\d+x-]+)"
            r"|u(?P<units>[\d+x-]+)"
            r"|s(?P<spells>[\d+x-]+)"
        )

        # Regex to parse an individual hero entry.
        # - hero_id is required.
        # - pet_id (prefixed by "p") is optional.
        # - Equipment (prefixed by "e") is optional; if present, eq1 is required, eq2 (after an underscore) is optional.
        hero_pattern = re.compile(
            r"(?P<hero_id>\d+)"
            r"(?:m\d+)?"
            r"(?:p(?P<pet_id>\d+))?"
            r"(?:e(?P<eq1>\d+)(?:_(?P<eq2>\d+))?)?"
        )

        # Iterate over all section matches.
        for match in ARMY_LINK_SEPARATOR.finditer(self.link):
            if match.group("heroes"):
                hero_entries = match.group("heroes").split('-')
                for hero in hero_entries:
                    m = hero_pattern.fullmatch(hero)
                    if m:
                        hero_id = HERO_BASE_ID + int(m.group("hero_id"))
                        pet_id = PET_BASE_ID + int(m.group("pet_id")) if m.group("pet_id") else None
                        eq1 = EQUIPMENT_BASE_ID + int(m.group("eq1")) if m.group("eq1") else None
                        eq2 = EQUIPMENT_BASE_ID + int(m.group("eq2")) if m.group("eq2") else None
                        self.heroes_loadout.append(HeroLoadout((hero_id, pet_id, eq1, eq2), self._lookup))

            elif match.group("castle_troops"):
                self._parse_items(match.group("castle_troops"), TROOP_BASE_ID, Troop, self.clan_castle_troops)

            elif match.group("castle_spells"):
                self._parse_items(match.group("castle_spells"), SPELL_BASE_ID, Spell, self.clan_castle_spells)

            elif match.group("units"):
                self._parse_items(match.group("units"), TROOP_BASE_ID, Troop, self.troops)

            elif match.group("spells"):
                self._parse_items(match.group("spells"), SPELL_BASE_ID, Spell, self.spells)

        self._lookup = None



from ._version import __version__

from .abc import BasePlayer, BaseClan
from .buildings import (
    Building,
    GearUp,
    MergeRequirement,
    SeasonalDefenseModule,
    SeasonalDefense,
    Supercharge,
    TownhallUnlock,
    TownhallWeapon,
    Trap,
)
from .characters import Guardian, Helper
from .clans import RankedClan, Clan
from .client import Client
from .constants import *
from .cosmetics import Skin, Scenery, Obstacle, Decoration, ClanCapitalHousePart
from .events import PlayerEvents, ClanEvents, WarEvents, EventsClient, ClientEvents
from .enums import (
    PlayerHouseElementType,
    Resource,
    Role,
    WarRound,
    WarState,
    BattleModifier,
    WarResult,
    ProductionBuildingType,
    BuildingType,
    VillageType,
    SceneryType,
    EquipmentRarity,
    SkinTier,
    ClanType
)
from .errors import (
    ClashOfClansException,
    HTTPException,
    NotFound,
    InvalidArgument,
    InvalidCredentials,
    Forbidden,
    Maintenance,
    GatewayError,
    PrivateWarLog,
)
from .game_data import AccountData, Upgrade, Boosts, ArmyRecipe, HeroLoadout, StaticData
from .hero import Equipment, Hero, Pet
from .http import BasicThrottler, BatchThrottler, HTTPClient
from .iterators import (
    ClanIterator,
    ClanWarIterator,
    PlayerIterator,
    LeagueWarIterator,
    CurrentWarIterator,
)
from .miscmodels import (
    Achievement,
    Badge,
    BaseLeague,
    CapitalDistrict,
    ChatLanguage,
    GoldPassSeason,
    Icon,
    Label,
    League,
    LegendStatistics,
    LoadGameData,
    Location,
    PlayerHouseElement,
    Season,
    Timestamp,
    TimeDelta,
    TID,
    Translation
)
from .players import Player, ClanMember, RankedPlayer
from .player_clan import PlayerClan
from .raid import RaidClan, RaidMember, RaidLogEntry, RaidDistrict, RaidAttack
from .spell import Spell
from .troop import Troop
from .war_clans import WarClan, ClanWarLeagueClan
from .war_attack import WarAttack
from .war_members import ClanWarLeagueClanMember, ClanWarMember
from .wars import ClanWar, ClanWarLogEntry, ClanWarLeagueGroup, ExtendedCWLGroup
from . import utils




from aiohttp import ClientResponse


class ClashOfClansException(Exception):
    """Base exception for clashy.py
    """


class HTTPException(ClashOfClansException):
    """Base exception for when a HTTP request fails

    Attributes
    ----------
    response:
        :class:`aiohttp.ClientResponse` - The response of the failed HTTP request.

    status:
        :class:`int` - The status code of the HTTP request

    reason:
        :class:`str` - The reason provided by the API.
            This could be an empty string if nothing was given.

    message:
        :class:`str` - The more detailed message provided by the API.
            This could be an empty string if nothing was given

    """

    __slots__ = ("response", "status", "message", "reason", "_data")

    def _from_response(self, response, data):
        if isinstance(response, int):
            self.status = response
            self.response = None
        elif isinstance(response, ClientResponse):
            self.response = response
            self.status = response.status
        else:
            self.response = self.status = None

        if isinstance(data, dict):
            self.reason = data.get("reason", "Unknown")
            self.message = data.get("message")
        elif isinstance(data, str):
            self.reason = data
            self.message = None
        else:
            self.reason = "Unknown"
            self.message = None

        fmt = "{0.reason} (status code: {0.status})"
        if self.message:
            fmt += ": {0.message}"

        super().__init__(fmt.format(self))

    def __init__(self, response=None, data=None):
        if isinstance(response, (ClientResponse, int)):
            self._from_response(response, data)
        else:
            self.response = None
            self.status = 0
            self.reason = None
            self.message = response

            fmt = "Error Occurred: {0}"
            super().__init__(fmt.format(self.message))


class InvalidArgument(HTTPException):
    """
    Thrown when an error status 400 occurs.

    Client provided incorrect parameters for the request.

    Subclass of :exc:`HTTPException`
    """


class InvalidCredentials(HTTPException):
    """Thrown when an error status 403 occurs and the reason is invalid credentials.

    Special Exception thrown when missing/incorrect credentials
    were passed. This is when your email/password pair is incorrect.
    Subclass of :exc:`HTTPException`
    """
    def __init__(self, response="Invalid Credentials"):
        super().__init__(response=response)


class Forbidden(HTTPException):
    """Thrown when an error status 403 occurs.

    API token does not grant access to the requested resource.

    Subclass of :exc:`HTTPException`"""


class PrivateWarLog(Forbidden):
    """Thrown when an error status 403 occurs when trying to get a clan's war info.

    This is a special subclass of :exc:`Forbidden` for when a clan's war log is private.
    """


class NotFound(HTTPException):
    """Thrown when an error status 404 occurs.

    The resource was not found.

    Subclass of :exc:`HTTPException`
    """


class Maintenance(HTTPException):
    """Thrown when an error status 503 occurs.

    Service is temporarily unavailable because of maintenance.

    Subclass of :exc:`HTTPException`
    """


class GatewayError(HTTPException):
    """Thrown when a gateway error occurs. These are either status 502 or 504

    Error code 502: Bad Gateway
    Error code 504: The Gateway has timed-out.

    Subclass of :exc:`HTTPException`
    """




from datetime import datetime
from typing import Any, Optional, Type, TypeVar

from .enums import ExtendedEnum, PlayerHouseElementType, VillageType
from .utils import from_timestamp

T = TypeVar("T")


def try_enum(_class: Type[T], data: Any, **kwargs) -> Optional[T]:
    """Helper function to create a class from the given data."""
    if issubclass(_class, ExtendedEnum):
        return data and _class(data)
    return data and _class(data=data, **kwargs)


class Achievement:
    """Represents a Clash of Clans Achievement.

    Attributes
    -----------
    name:
        :class:`str`: The name of the achievement.
    stars:
        :class:`int`: The current stars achieved for the achievement.
    value:
        :class:`int`: The number of X things attained for this achievement.
    target:
        :class:`int`: The number of X things required to complete this achievement.
    info:
        :class:`str`: Information regarding the achievement.
    completion_info:
        :class:`str`: Information regarding completion of the achievement.
    village:
        :class:`str`: Either ``home`` or ``builderBase``.
    """

    __slots__ = (
        "name",
        "stars",
        "value",
        "target",
        "info",
        "completion_info",
        "village",
    )

    def __str__(self):
        return self.name

    def __repr__(self):
        attrs = [
            ("name", self.name),
            ("stars", self.stars),
            ("value", self.value),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __init__(self, *, data):
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        self.name: str = data["name"]
        self.stars: int = data["stars"]
        self.value: int = data["value"]
        self.target: int = data["target"]
        self.info: str = data["info"]
        self.completion_info: str = data["completionInfo"]
        self.village = VillageType(value=data["village"])

    @property
    def is_builder_base(self) -> bool:
        """:class:`bool`: Returns a boolean which indicates if the achievement belongs to the builder base"""
        return self.village == VillageType.builder_base

    @property
    def is_home_base(self) -> bool:
        """:class:`bool`: Returns a boolean which indicates if the achievement belongs to the home base"""
        return self.village == VillageType.home

    @property
    def is_clan_capital(self) -> bool:
        """:class:`bool`: Returns a boolean which indicates if the achievement belongs to the clan capital"""
        return self.village == VillageType.clan_capital

    @property
    def is_completed(self) -> bool:
        """:class:`bool`: Returns a boolean which indicates whether the achievement is completed (3 stars achieved)"""
        return self.stars == 3


class LoadGameData:
    """Pass this into the ``load_game_data`` parameter of :class:`Client`.

    See :ref:`game_data` for more information.

    Parameters
    ----------
    always: bool
        Whether to always inject game metadata into objects.

    default: bool
        Always inject game metadata into objects, except when running events tasks.

    startup_only: bool
        Never automatically inject game metadata into objects, but load it up on startup regardless for use with
        the ``load_game_data`` parameter of :meth:`Client.get_player` or :meth:`Client.parse_army_link`.

    never: bool
        Never inject game metadata, and don't load it on startup.

    """
    always = False
    default = False
    startup_only = False
    never = False

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            try:
                getattr(self.__class__, key)
            except AttributeError:
                raise RuntimeError("%s was not a valid LoadGameData option.", key)
            else:
                setattr(self.__class__, key, value)


class TimeDelta:
    """Represents a Timedelta object corresponding to things that take time to do in the API.

    Some examples include:

    * Upgrade times
    * Training times
    * Cooldown times

    This object works in a very similar fashion to datetime's `timedelta <https://docs.python.org/3/library/datetime.html#timedelta-objects>`_
    object, but with a few more helpful attributes.

    .. note::

        You should not construct this yourself, instead use it from the attribute of an e.g. :class:`Troop` model.


    Attributes
    ----------
    days: int
        The number of days in the timedelta.
    hours: int
        The number of hours in the timedelta. This does not include days.
        For example, if an upgrade took 36 hours, the ``.days`` attribute would be 1, and ``.hours`` would be 12.
    minutes: int
        The number of minutes in the timedelta. The same logic applies as with hours.
    seconds: int
        The number of seconds in the timedelta. The same logic applies as with hours.

    """
    def __init__(self, days=0, hours=0, minutes=0, seconds=0):
        _days, _hours = divmod(hours, 24)
        _hours_left, _mins = divmod(minutes, 60)

        self.days = days + _days
        self.hours = hours + _hours + _hours_left
        self.minutes = minutes + _mins
        self.seconds = seconds

    def total_seconds(self):
        """Returns the total number of seconds in the time object.

        This is the addition of all days, hours, minutes, seconds.

        Returns
        -------
        int
            The number of seconds"""
        return self.days * 24 * 60 * 60 + \
               self.hours * 60 * 60 + \
               self.minutes * 60 + \
               self.seconds


class Location:
    """Represents a Clash of Clans Location

    Attributes
    -----------
    id:
        :class:`int` - The location ID
    name:
        :class:`str` - The location name
    is_country:
        :class:`bool` - Indicates whether the location is a country
    country_code:
        :class:`str` - The shorthand country code, if the location is a country
    localised_name:
        :class:`str` - A localised name of the location. The extent of the use of this is unknown at present.
    """

    __slots__ = ("id", "name", "is_country", "country_code", "localised_name")

    def __str__(self):
        return self.name

    def __repr__(self):
        attrs = [
            ("id", self.id),
            ("name", self.name),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def __init__(self, *, data):
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        # pylint: disable=invalid-name
        data_get = data.get

        self.id: int = data_get("id")
        self.name: str = data_get("name")
        self.is_country: bool = data_get("isCountry")
        self.country_code: str = data_get("countryCode")
        self.localised_name: str = data_get("localizedName")


class BaseLeague:
    """Represents a basic league.

    Attributes
    -----------
    id: :class:`int`: The league's unique ID
    name: :class:`str`: The league's name, as it appears in-game."""

    __slots__ = (
        "id",
        "name",
        "_client"
    )

    def __init__(self, *, data, client=None):
        # pylint: disable=invalid-name
        self.id: int = data["id"]
        self.name: str = data["name"]
        self._client = client

    def __repr__(self):
        return "<%s id=%s name=%s>" % (self.__class__.__name__, self.id, self.name)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return isinstance(self, other.__class__) and other.id == self.id


class League(BaseLeague):
    """Represents a Clash of Clans League

    Attributes
    -----------
    id:
        :class:`int`: The league ID.
    name:
        :class:`str`: The league name.
    icon:
        :class:`Icon`: The league's icon.
    """

    __slots__ = (
        "icon",
    )

    def __str__(self):
        return self.name

    def __repr__(self):
        attrs = [("id", self.id), ("name", self.name)]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def __init__(self, *, data, client):
        super().__init__(data=data, client=client)
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        # pylint: disable=invalid-name
        data_get = data.get
        self.icon = try_enum(Icon, data=data_get("iconUrls"), client=self._client)


class Season:
    """Represents a Clash of Clans Player's Season."""

    # pylint: disable=invalid-name

    __slots__ = ("rank", "trophies", "id")

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.rank == other.rank
            and self.trophies == other.trophies
            and self.id == other.id
        )

    def __init__(self, *, data):
        self.rank: int = data.get("rank")
        self.trophies: int = data.get("trophies")
        self.id: int = data.get("id")


class LegendStatistics:
    """Represents the Legend Statistics for a player.

    Attributes
    -----------
    legend_trophies:
        :class:`int` - The player's legend trophies
    current_season:
        :class:`Season`: Legend statistics for this season.
    previous_season:
        :class:`Season`: Legend statistics for the previous season.
    best_season:
        :class:`Season`: Legend statistics for the player's best season.
    previous_builder_base_season:
        :class:`Season`: Legend statistics for the previous builder base season.
    best_builder_base_season:
        :class:`Season`: Legend statistics for the player's best builder base season.
    """

    __slots__ = ("legend_trophies", "current_season", "previous_season", "best_season", "previous_builder_base_season",
                 "best_builder_base_season")

    def __repr__(self):
        attrs = [
            ("legend_trophies", self.legend_trophies),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.best_season == other.best_season
            and self.current_season == other.current_season
            and self.best_builder_base_season == other.best_builder_base_season
            and self.previous_season == other.previous_season
            and self.previous_builder_base_season == self.previous_builder_base_season
        )

    def __init__(self, *, data):
        self.legend_trophies: int = data["legendTrophies"]
        self.current_season = try_enum(Season, data=data.get("currentSeason"))
        self.previous_season = try_enum(Season, data=data.get("previousSeason"))
        self.best_season = try_enum(Season, data=data.get("bestSeason"))
        self.previous_builder_base_season = try_enum(Season, data=data.get("previousBuilderBaseSeason"))
        self.best_builder_base_season = try_enum(Season, data=data.get("bestBuilderBaseSeason"))


class Badge:
    """Represents a Clash Of Clans Badge.

    Attributes
    -----------
    small:
        :class:`str` - URL for a small sized badge (70x70).
    medium:
        :class:`str` - URL for a medium sized badge (200x200).
    large:
        :class:`str` - URL for a large sized badge (512x512).
    """

    __slots__ = ("small", "medium", "large", "_client")

    def __repr__(self):
        attrs = [
            ("url", self.url),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __init__(self, *, data, client):
        self._client = client

        self.small: str = data.get("small")
        self.medium: str = data.get("medium")
        self.large: str = data.get("large")

    @property
    def url(self) -> str:
        """:class:`str`: the default icon URL. Returns the medium-sized icon URL if available.
        Falls back to small and large (in that order) if not"""
        return self.medium or self.small or self.large

    async def save(self, filepath, size=None) -> int:
        """
        Save this badge as a file-like object.

        Parameters
        -----------
        filepath: :class:`os.PathLike`
            The filename to save the badge to.
        size: Optional[:class:`str`]
            Either ``small``, ``medium`` or ``large``. The default is ``medium``.

        Returns
        --------
        The number of bytes written: :class:`int`

        Raises
        ------
        :exc:`HTTPException`:
            Saving the badge failed.
        :exc:`NotFound`:
            The URL was not found.
        """
        sizes = {"small": self.small, "medium": self.medium, "large": self.large}

        if size and size in sizes.keys():
            url = sizes[size]
        else:
            url = self.url

        data = await self._client.http.get_data_from_url(url)

        with open(filepath, "wb") as file:
            return file.write(data)


class Icon:
    """Represents a Clash Of Clans Icon.

    Attributes
    -----------
    tiny:
        :class:`str`: URL for a tiny sized icon (32x32).
    small:
        :class:`str`: URL for a small sized icon (72x72).
    medium:
        :class:`str`: URL for a medium sized icon (288x288).
    """

    __slots__ = ("small", "medium", "tiny", "_client")

    def __repr__(self):
        attrs = [
            ("url", self.url),
        ]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __init__(self, *, data, client):
        self._client = client

        self.tiny: str = data.get("tiny")
        self.small: str = data.get("small")
        self.medium: str = data.get("medium")

    @property
    def url(self) -> str:
        """:class:`str`: the default icon URL. Returns the medium-sized icon URL if available.
        Falls back to small and tiny (in that order) if not"""
        return self.medium or self.small or self.tiny

    async def save(self, filepath: str, size: Optional[str] = None) -> int:
        """
        Save this icon as a file-like object.

        Parameters
        -----------
        filepath: :class:`os.PathLike`
            The filename to save the badge to.
        size: Optional[:class:`str`]
            Either ``tiny``, ``small`` or ``medium``. The default is ``small``.

        Returns
        --------
        :class:`int`: The number of bytes written.

        Raises
        ------
        :exc:`HTTPException`:
            Saving the badge failed.
        :exc:`NotFound`:
            The URL was not found.
        """
        sizes = {"tiny": self.tiny, "small": self.small, "medium": self.medium}

        if size and size in sizes.keys():
            url = sizes[size]
        else:
            url = self.url

        data = await self._client.http.get_data_from_url(url)

        with open(filepath, "wb") as file:
            return file.write(data)


class Timestamp:
    """Represents a Clash of Clans Timestamp

    Attributes
    -----------
    raw_time:
        :class:`str`: The raw timestamp string (ISO8601) as given by the API.
    """

    __slots__ = ("raw_time", "_data")

    def __repr__(self):
        attrs = [("time", self.time), ("seconds_until", self.seconds_until)]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.raw_time == other.raw_time

    def __lt__(self, other):
        if not isinstance(other, Timestamp) or not isinstance(self, Timestamp):
            return NotImplemented

        return self.time < other.time

    def __le__(self, other):
        less_than = Timestamp.__lt__(other, self)
        if less_than is NotImplemented:
            return NotImplemented
        return not less_than

    def __init__(self, *, data):
        self.raw_time = data

    @property
    def time(self) -> datetime:
        """:class:`datetime`: Returns the timestamp as a UTC datetime object."""
        return from_timestamp(self.raw_time)

    @property
    def now(self) -> datetime:
        """:class:`datetime`: Returns the time of the timestamp as a datetime object in UTC."""
        return datetime.utcnow()

    @property
    def seconds_until(self) -> int:
        """:class:`int`: Returns the number of seconds until the timestamp. This may be negative."""
        delta = self.time - self.now
        return int(delta.total_seconds())


class Label:
    """Represents a clan or player label.

    Attributes
    -----------
    id:
        :class:`int`: The label's unique ID as given by the API.
    name:
        :class:`str`: The label's name.
    badge:
        :class:`Badge`: The label's badge.
    """

    __slots__ = ("id", "name", "_client", "badge")

    def __str__(self):
        return self.name

    def __repr__(self):
        attrs = [("id", self.id), ("name", self.name)]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def __init__(self, *, data, client):
        # pylint: disable=invalid-name
        self._client = client

        self.id: int = data.get("id")
        self.name: str = data.get("name")
        self.badge = try_enum(Icon, data=data.get("iconUrls"), client=self._client)


class CapitalDistrict:
    """Represents a Clan Capital District.

    Attributes
    -----------
    id:
        :class:`int`: The district's unique ID as given by the API.
    name:
        :class:`str`: The district's name.
    hall_level:
        :class:`int`: The district's hall level
    """

    __slots__ = ("id", "name", "hall_level")

    def __str__(self):
        return self.name

    def __repr__(self):
        attrs = [("id", self.id), ("name", self.name)]
        return "<%s %s>" % (self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
               self.id == other.id and \
               self.hall_level == other.hall_level

    def __init__(self, *, data, client):
        # pylint: disable=invalid-name
        self.id: int = data.get("id")
        self.name: str = data.get("name")
        self.hall_level: int = data.get("districtHallLevel")


class ChatLanguage:
    """Represents a clan's chat language.

    Attributes
    ----------
    id: :class:`int`: The language's unique ID
    name: :class:`str`: The language's full name, for example ``English``.
    language_code: :class:`str` The language's abbreviated code, for example ``EN``.
    """
    __slots__ = (
        "id",
        "name",
        "language_code"
    )

    def __init__(self, *, data):
        # pylint: disable=invalid-name
        self.id: int = data["id"]
        self.name: str = data["name"]
        self.language_code: str = data["languageCode"]

    def __repr__(self):
        return "<%s id=%s name=%s>" % (self.__class__.__name__, self.id, self.name)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return isinstance(self, other.__class__) and other.id == self.id


class GoldPassSeason:
    """Represents a gold pass season.

    Attributes
    ----------
    start_time:
        :class:`Timestamp`: The gold pass season start time
    end_time:
        :class:`Timestamp`: The gold pass season end time
    duration:
        :class:`datetime.timedelta`: The duration of the gold pass season
    """
    __slots__ = ("start_time", "end_time", "duration")

    def __init__(self, *, data):
        self.start_time = try_enum(Timestamp, data.get("startTime"))
        self.end_time = try_enum(Timestamp, data.get("endTime"))
        self.duration = self.end_time.time - self.start_time.time

    def __eq__(self, other):
        return (isinstance(other, GoldPassSeason)
                and self.start_time == other.start_time
                and self.end_time == other.end_time)


class PlayerHouseElement:
    """Represents an element of a player house.

    Attributes
    ----------
    id:
        :class:`int`: The id of the house element
    type:
        :class:`PlayerHouseElementType`: The type of the house element
    """
    __slots__ = ("id", "type")

    def __init__(self, *, data):
        self.id = data.get("id")
        self.type = data.get("type") and PlayerHouseElementType(value=data["type"])

    def __eq__(self, other):
        return (isinstance(other, PlayerHouseElement)
                and self.id == other.id
                and self.type == other.type)


class TID:
    """Represents a Translation ID (TID) for Clash of Clans game elements.
    
    Attributes
    ----------
    name: :class:`str`
        The translation ID for the name of the game element.
    info: :class:`str`
        The translation ID for additional info/description of the game element.
    """
    
    __slots__ = ("name", "info")
    
    def __init__(self, data: dict):
        self.name = data["name"]
        self.info = data.get("info", "")


class Translation:
    """Represents translations for multiple languages.
    
    Supports multiple access patterns:
    
    * Attribute access: ``translation.russian``, ``translation.english``
    * Dictionary access (uppercase): ``translation["RU"]``, ``translation["EN"]``
    * Dictionary access (lowercase): ``translation["ru"]``, ``translation["en"]``
    
    Attributes
    ----------
    english: :class:`str`
    arabic: :class:`str`
    chinese: :class:`str`
    chinese_traditional: :class:`str`
    german: :class:`str`
    spanish: :class:`str`
    persian: :class:`str`
    finnish: :class:`str`
    french: :class:`str`
    indonesian: :class:`str`
    italian: :class:`str`
    japanese: :class:`str`
    korean: :class:`str`
    malay: :class:`str`
    dutch: :class:`str`
    norwegian: :class:`str`
    polish: :class:`str`
    portuguese: :class:`str`
    russian: :class:`str`
    thai: :class:`str`
    turkish: :class:`str`
    vietnamese: :class:`str`
    """
    
    __slots__ = (
        "english", "arabic", "chinese", "chinese_traditional", "german",
        "spanish", "persian", "finnish", "french", "indonesian", "italian",
        "japanese", "korean", "malay", "dutch", "norwegian", "polish",
        "portuguese", "russian", "thai", "turkish", "vietnamese"
    )
    
    _LANGUAGE_MAP = {
        "EN": "english",
        "AR": "arabic",
        "CN": "chinese",
        "CNT": "chinese_traditional",
        "DE": "german",
        "ES": "spanish",
        "FA": "persian",
        "FI": "finnish",
        "FR": "french",
        "ID": "indonesian",
        "IT": "italian",
        "JP": "japanese",
        "KR": "korean",
        "MS": "malay",
        "NL": "dutch",
        "NO": "norwegian",
        "PL": "polish",
        "PT": "portuguese",
        "RU": "russian",
        "TH": "thai",
        "TR": "turkish",
        "VI": "vietnamese"
    }
    
    def __init__(self, data: dict):
        """Initialize Translation from a dictionary with language codes as keys.
        
        Parameters
        ----------
        data: :class:`dict`
            Dictionary with language codes (EN, AR, CN, etc.) as keys and translations as values.
        """
        self.english: str = data.get("EN", "")
        self.arabic: str = data.get("AR", "")
        self.chinese: str = data.get("CN", "")
        self.chinese_traditional: str = data.get("CNT", "")
        self.german: str = data.get("DE", "")
        self.spanish: str = data.get("ES", "")
        self.persian: str = data.get("FA", "")
        self.finnish: str = data.get("FI", "")
        self.french: str = data.get("FR", "")
        self.indonesian: str = data.get("ID", "")
        self.italian: str = data.get("IT", "")
        self.japanese: str = data.get("JP", "")
        self.korean: str = data.get("KR", "")
        self.malay: str = data.get("MS", "")
        self.dutch: str = data.get("NL", "")
        self.norwegian: str = data.get("NO", "")
        self.polish: str = data.get("PL", "")
        self.portuguese: str = data.get("PT", "")
        self.russian: str = data.get("RU", "")
        self.thai: str = data.get("TH", "")
        self.turkish: str = data.get("TR", "")
        self.vietnamese: str = data.get("VI", "")
    
    def __getitem__(self, key: str) -> str:
        """Get translation by language code (case-insensitive).
        
        Parameters
        ----------
        key: :class:`str`
            Language code (e.g., "EN", "en", "RU", "ru")
        
        Returns
        -------
        :class:`str`
            The translation for the specified language.
        
        Raises
        ------
        KeyError
            If the language code is not supported.
        """
        key_upper = key.upper()
        if key_upper in self._LANGUAGE_MAP:
            return getattr(self, self._LANGUAGE_MAP[key_upper])
        raise KeyError(f"Language code '{key}' not supported")
    
    def __repr__(self):
        return f"<Translation english={self.english!r}>"
    
    def __eq__(self, other):
        return isinstance(other, Translation) and all(
            getattr(self, attr) == getattr(other, attr)
            for attr in self.__slots__
        )






# Enables circular import for type hinting coc.Client
from __future__ import annotations

import itertools

from typing import AsyncIterator, List, Optional, Type, TYPE_CHECKING, Union

from .enums import BattleModifier, WarResult, WarRound, WarState
from .iterators import LeagueWarIterator
from .miscmodels import try_enum, Timestamp, TID
from .utils import cached_property, get
from .war_clans import WarClan, ClanWarLeagueClan
from .war_attack import WarAttack

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from .war_members import ClanWarMember  # noqa

class ClanWar:
    """Represents a Current Clash of Clans War

    Attributes
    ----------
    clan:
        :class:`WarClan`: The home clan.
    opponent:
        :class:`WarClan`: The opposition clan.
    state:
        :class:`WarState`: The clan's current war state.
    preparation_start_time:
        :class:`Timestamp`: The :class:`Timestamp` that preparation day started at.
    start_time:
        :class:`Timestamp`: The :class:`Timestamp` that battle day starts at.
    end_time:
        :class:`Timestamp`: The :class:`Timestamp` that battle day ends at.
    team_size:
        :class:`int`: The number of players on each side of the war.
    war_tag:
        :class:`str`: The war's unique tag. This is ``None`` unless this is a Clan League War (CWL).
    league_group:
        :class:`ClanWarLeagueGroup`: The war's league group. This is ``None`` unless this is a Clan League War.
    attacks_per_member:
        :class:`int`: The number of attacks each member has this war.
    battle_modifier:
        :class:`BattleModifier`: The battle modifier for this war.
    clan_cls: :class:`WarClan`
        the type `war.clan` and `war.opponent` will be of.
        Ensure any overriding of this inherits from :class:`coc.WarClan`.
    """

    __slots__ = (
        "state",
        "preparation_start_time",
        "start_time",
        "end_time",
        "team_size",
        "war_tag",
        "_client",
        "clan_tag",
        "clan",
        "opponent",
        "league_group",
        "attacks_per_member",
        "clan_cls",
        "_response_retry",
        "_raw_data",
        "battle_modifier",
    )

    def __init__(self, *, data, client, **kwargs):
        self._response_retry = data.get("_response_retry")
        self._client = client
        self._raw_data = data if client and client.raw_attribute else None
        self.clan_tag = kwargs.pop("clan_tag", None)
        self.clan_cls = kwargs.pop('clan_cls', WarClan)
        self._from_data(data)

        self.clan_tag = self.clan and self.clan.tag or self.clan_tag
        self.league_group: ClanWarLeagueGroup | None = kwargs.pop("league_group", None)

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.state: WarState = try_enum(WarState, data=data_get("state"))
        self.preparation_start_time = try_enum(Timestamp, data=data_get(
            "preparationStartTime"))
        self.start_time = try_enum(Timestamp, data=data_get("startTime"))
        self.end_time = try_enum(Timestamp, data=data_get("endTime"))
        self.war_tag: str = data_get("tag")
        self.battle_modifier: BattleModifier = try_enum(BattleModifier, data=data_get('battleModifier', 'none'))
        if data_get("attacksPerMember") is None or self.is_cwl:
            self.attacks_per_member: int = 1
        else:
            self.attacks_per_member: int = data_get("attacksPerMember")

        self.team_size: int = data_get("teamSize") or len(
            data_get("clan", {}).get("members", []))

        clan_data = data_get("clan")
        # annoying bug where if you request a war with a clan tag that clan could be the opponent or clan,
        # depending on the way the game stores it internally. This isn't very helpful as we always want it
        # from the perspective of the tag we provided, so switch them around if it isn't correct.
        if clan_data and clan_data.get("tag", self.clan_tag) == self.clan_tag:
            self.clan = try_enum(self.clan_cls, data=clan_data, client=self._client,
                                 war=self)
            self.opponent = try_enum(self.clan_cls, data=data_get("opponent"),
                                     client=self._client, war=self)
        else:
            self.clan = try_enum(self.clan_cls, data=data_get("opponent"),
                                 client=self._client, war=self)
            self.opponent = try_enum(self.clan_cls, data=clan_data,
                                     client=self._client, war=self)

    def __eq__(self, other):
        if not isinstance(other, ClanWar):
            return NotImplemented
        return {self.clan.tag, self.opponent.tag} == {other.clan.tag, other.opponent.tag} and \
            self.preparation_start_time == other.preparation_start_time


    @property
    def attacks(self) -> List[WarAttack]:
        """List[:class:`WarAttack`]: Returns all attacks this war, sorted by attack order."""
        return sorted([*self.clan.attacks, *self.opponent.attacks],
                      key=lambda x: x.order, reverse=True)

    @property
    def members(self) -> List["ClanWarMember"]:
        """List[:class:`ClanWarMember`]: A list of members that are in the war."""
        return sorted([*self.clan.members, *self.opponent.members],
                      key=lambda x: (not x.is_opponent, x.map_position))

    @property
    def type(self) -> Optional[str]:
        """:class:`str`: Returns either ``friendly``, ``random`` or ``cwl``.

        This will returns ``None`` if the clan is not in war, or ``cwl`` if the clan is in a league war.

        Possibilities for the length of preparation time for a friendly war include:
        5 minutes, 15 minutes, 30 minutes, 1 hour, 2 hours, 4 hours, 6 hours, 8 hours, 12 hours,
        16 hours, 20 hours or 24 hours.
        """
        if self.war_tag:
            return "cwl"

        if not self.start_time:
            return None

        prep_list = [
            5 * 60,
            15 * 60,
            30 * 60,
            60 * 60,
            2 * 60 * 60,
            4 * 60 * 60,
            6 * 60 * 60,
            8 * 60 * 60,
            12 * 60 * 60,
            16 * 60 * 60,
            20 * 60 * 60,
            24 * 60 * 60,
        ]
        if (
                self.start_time.time - self.preparation_start_time.time).seconds in prep_list:
            return "friendly"

        return "random"

    @property
    def status(self) -> str:
        """:class:`str`: Returns the war status, based off the home clan.

        Strings returned are determined by result and state, as listed below:

        +------------+-------------+
        | ``inWar``  | ``warEnded``|
        +------------+-------------+
        | ``winning``| ``won``     |
        +------------+-------------+
        | ``tied``   | ``tie``     |
        +------------+-------------+
        | ``losing`` | ``lost``    |
        +------------+-------------+
        """
        # pylint: disable=too-many-return-statements
        if self.state == "inWar":
            if self.clan.stars > self.opponent.stars:
                return "winning"

            if self.clan.stars == self.opponent.stars:
                if self.clan.destruction > self.opponent.destruction:
                    return "winning"
                if self.clan.destruction == self.opponent.destruction:
                    return "tied"

            return "losing"

        if self.state == "warEnded":
            if self.clan.stars > self.opponent.stars:
                return "won"

            if self.clan.stars == self.opponent.stars:
                if self.clan.destruction > self.opponent.destruction:
                    return "won"
                if self.clan.destruction == self.opponent.destruction:
                    return "tie"

            return "lost"

        return ""

    @property
    def is_cwl(self) -> bool:
        """:class:`bool`: Returns a boolean indicating if the war is a Clan War League (CWL) war."""
        return self.type == "cwl"

    def get_member(self, tag: str) -> Optional["ClanWarMember"]:
        """Return a :class:`ClanWarMember` with the tag provided. Returns ``None`` if not found.

        Example
        --------
        .. code-block:: python3

            war = await client.get_current_war('clan_tag')
            member = war.get_member('player_tag')

        Returns
        --------
        Optional[:class:`ClanWarMember`]: The member who matches the tag provided.
        """

        home_member = self.clan.get_member(tag)
        if home_member:
            return home_member

        away_member = self.opponent.get_member(tag)
        return away_member

    def get_member_by(self, **attrs) -> Optional["ClanWarMember"]:
        """Returns the first :class:`WarMember` that meets the attributes passed

        This will return the first member matching the attributes passed.

        An example of this looks like:

        .. code-block:: python3

            member = ClanWar.get_member(tag='tag')

        This search implements the :func:`coc.utils.get` function
        """
        return get(self.members, **attrs)

    def get_attack(self, attacker_tag: str, defender_tag: str) -> Optional[
        WarAttack]:
        """Return the :class:`WarAttack` with the attacker tag and defender tag provided.

        If the attack was not found, this will return ``None``.

        Returns
        --------
        The attack with the correct attacker and defender tags: :class:`WarAttack`: """
        attacker = self.get_member(attacker_tag)
        if not attacker:
            return None

        attacks = attacker.attacks
        if len(attacks) == 0:
            return None
        return get(attacks, defender_tag=defender_tag)

    def get_defenses(self, defender_tag: str) -> List[WarAttack]:
        """Return a :class:`list` of :class:`WarAttack` for the defender tag provided.

        If the player has no defenses, this will return an empty list.

        Returns
        -------
        The player's defenses: List[:class:`WarAttack`]"""
        defender = self.get_member(defender_tag)
        # we could do a global lookup on all attacks in the war but this is faster as we have to lookup half the attacks
        if defender.is_opponent:
            # we need to get home clan's attacks on this base
            return list(filter(lambda x: x.defender_tag == defender_tag,
                               self.clan.attacks))

        return list(filter(lambda x: x.defender_tag == defender_tag,
                           self.opponent.attacks))


class ClanWarLogEntry:
    """Represents a Clash of Clans War Log Entry

    .. note::

        Please see the :class:`WarClan` documentation for a full list of missing attributes,
        as the clan and opponent attributes are only partially filled by the API.

        If the :attr:`ClanWarLogEntry.type` is ``cwl``, the :attr:`WarClan.attack_count`, :attr:`WarClan.stars`
        and :attr:`WarClan.destruction` are all a total which over the period of that CWL season.

        In addition, if it is a CWL entry, ``opponent`` and ``result`` will be ``None``.


    Attributes
    ----------
    result:
        :class:`WarResult`: The result of the war - ``win``, ``lose`` or ``tie``
    end_time:
        :class:`Timestamp`: The :class:`Timestamp` that the war ended at.
    team_size:
        :class:`int`: The number of players on each side of the war.
    clan:
        :class:`WarClan`: The home clan.
    opponent:
        :class:`WarClan`: The opposition clan.
    attacks_per_member:
        :class:`int`: The number of attacks each member had this war.
    battle_modifier:
        :class:`BattleModifier`: The battle modifier for this war.
    """

    __slots__ = (
        "result", "end_time", "team_size", "clan", "opponent", "_client",
        "attacks_per_member", "battle_modifier", "_raw_data", "_response_retry")

    def __init__(self, *, data, client, **kwargs):
        self._client = client
        self._raw_data = data if client and client.raw_attribute else None
        self._from_data(data)
        self._response_retry = kwargs['response_retry'] if "response_retry" in kwargs else 0

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            if self.clan == other.clan \
                    and self.opponent == other.opponent \
                    and self.result == other.result \
                    and self.end_time == other.end_time \
                    and self.attacks_per_member == other.attacks_per_member:
                return True

        return False

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.result: WarResult = try_enum(WarResult, data=data_get("result"))
        self.end_time = try_enum(Timestamp, data=data_get("endTime"))
        self.team_size: int = data_get("teamSize")

        self.clan = self._fake_load_clan(data_get("clan"))
        self.opponent = self._fake_load_clan(data_get("opponent"))
        self.battle_modifier: BattleModifier = try_enum(BattleModifier, data=data_get('battleModifier', 'none'))

        if data_get("attacksPerMember") is None and self.is_league_entry:
            self.attacks_per_member: int = 1
        else:
            self.attacks_per_member: int = data_get("attacksPerMember")

    def _fake_load_clan(self, data):
        if not (data and data.get(
                "tag")):  # CWL seasons have an opposition with only badges and no tag/name.
            return None

        data["teamSize"] = self.team_size
        return try_enum(WarClan, data=data, client=self._client, war=None)

    @property
    def is_league_entry(self) -> bool:
        """:class:`bool`: Boolean indicating if the entry is a Clan War League (CWL) entry."""
        return self.result is None or self.opponent is None


class ClanWarLeagueGroup:
    """Represents a Clan War League (CWL) Group

    Attributes
    ----------
    state:
        :class:`str`: The current state of the league group (`inWar`, `preparation` etc.)
    season:
        :class:`str`: The current season of the league group
    number_of_rounds:
        :class:`int`: The number of rounds this league group contains.
    rounds:
        List[List[:class:`str`]]: A list of lists containing all war tags for each round.

        .. note::

            This only returns the current or past rounds. Any future rounds filled with #0 war tags will not appear.

            To find the number of rounds in this season, use :attr:`LeagueGroup.number_of_rounds`.

    """

    __slots__ = (
        "state", "season", "rounds", "number_of_rounds", "_client",
        "__iter_clans",
        "_cs_clans", "_raw_data")

    def __repr__(self):
        attrs = [
            ("state", self.state),
            ("season", self.season),
        ]
        return "<%s %s>" % (
            self.__class__.__name__, " ".join("%s=%r" % t for t in attrs),)

    def __init__(self, *, data, client, **_):
        self._client = client
        self._raw_data = data if client and client.raw_attribute else None
        self._from_data(data)

    def _from_data(self, data: dict) -> None:
        data_get = data.get

        self.state: str = data_get("state")
        self.season: str = data_get("season")

        rounds = data_get("rounds", [])
        self.number_of_rounds: int = len(rounds)
        # the API returns a list and the rounds that haven't started contain war tags of #0 (not sure why)...
        # we want to get only the valid rounds
        self.rounds: List[List[str]] = [n["warTags"] for n in rounds if
                                        n["warTags"][0] != "#0"]

        self.__iter_clans = (ClanWarLeagueClan(data=data, client=self._client)
                             for data in data_get("clans", []))

    @cached_property("_cs_clans")
    def clans(self) -> List[ClanWarLeagueClan]:
        """List[:class:`LeagueClan`]: Returns all participating clans."""
        return list(self.__iter_clans)

    def get_wars_for_clan(self, clan_tag: str, cls: Type[ClanWar] = ClanWar) -> \
            AsyncIterator[ClanWar]:
        """Returns every war the clan has participated in this current CWL.

        This returns a :class:`LeagueWarIterator` which fetches all wars in parallel.

        Example
        --------

        .. code-block:: python3

            group = await client.get_league_group('#clan_tag')

            async for war in group.get_wars_for_clan('#clantag'):
                print(war.start_time)

        Parameters
        ----------
        clan_tag: str
            The clan tag to get wars for. This method will only return wars which belong to this clan.
        cls: Type[:class:`ClanWar`]: The constructor used to create the league war.
                                     This should inherit from :class:`ClanWar`.

        Yields
        ------
        :class:`ClanWar`
            A war in the current CWL season with the clan in it..
        """
        return LeagueWarIterator(client=self._client,
                                 tags=itertools.chain(*self.rounds),
                                 clan_tag=clan_tag, cls=cls)

    def get_wars(
            self, cwl_round: WarRound = WarRound.current_war,
            cls: Type[ClanWar] = ClanWar
    ) -> AsyncIterator[ClanWar]:
        """Returns war information for every war in a league round.

        This returns a :class:`LeagueWarIterator` which fetches all wars in parallel.

        Example
        --------

        .. code-block:: python3

            group = await client.get_league_group('clan_tag')

            async for war in group.get_wars():
                print(war.clan_tag)

        Parameters
        ----------

        cls: Type[:class:`ClanWar`]: The constructor used to create the league war.
                                     This should inherit from :class:`ClanWar`.
        cwl_round: :class:`WarRound`
            An enum detailing the type of round to get. Could be ``coc.WarRound.previous_war``,
            ``coc.WarRound.current_war`` or ``coc.WarRound.preparation``.
            This defaults to ``coc.WarRound.current_war``.


        Yields
        ------
        :class:`ClanWar`
            A war in the requested round.
        """
        is_prep = self.state == "preparation"
        num_rounds = len(self.rounds)
        if cwl_round is WarRound.current_war and is_prep:
            round_tags = ()  # for round 1 and 15min prep between rounds this is a shortcut.
        elif cwl_round is WarRound.current_preparation and self.state == "warEnded":
            round_tags = ()  # for the end of CWL there's no next prep day.
        elif cwl_round is WarRound.previous_war and num_rounds == 1:
            round_tags = ()  # no previous war for first rounds.
        elif cwl_round is WarRound.previous_war and is_prep:
            round_tags = self.rounds[-2]
        elif cwl_round is WarRound.previous_war:
            round_tags = self.rounds[-3]
        elif cwl_round is WarRound.current_war:
            round_tags = self.rounds[-2]
        elif cwl_round is WarRound.current_preparation:
            round_tags = self.rounds[-1]
        else:
            round_tags = ()

        return LeagueWarIterator(client=self._client, tags=round_tags, cls=cls)


class ExtendedCWLGroup:
    def __init__(self, data: dict):
        self._id: int = data["id"]
        self.name: str = data["name"]
        self.TID = TID(data=data["TID"])
        self.first_place_medals: int = data["cwl_medals"]["first_place"]
        self.positional_medal_difference: int = data["cwl_medals"]["position_medal_diff"]
        self.bonus_medals: int = data["cwl_medals"]["bonus_reward"]
        self.minimum_number_of_bonuses: int = data["cwl_medals"]["minimum_bonus_amount"]
        self.promotions: int = data["promotions"]
        self.demotions: int = data["demotions"]
        self.only15v15: bool = data["only15v15"]



from coc.abc import BaseClan


class PlayerClan(BaseClan):
    """Represents a clan that belongs to a player.

    Attributes
    ----------
    tag: :class:`str`
        The clan's tag
    name: :class:`str`
        The clan's name
    badge: :class:`Badge`
        The clan's badge
    level: :class:`int`
        The clan's level.
    """




from .abc import BaseDataClass, LeveledUnit, LevelManager
from .enums import Resource, VillageType, BuildingType
from .miscmodels import TID, TimeDelta


class SeasonalDefenseModule(LeveledUnit):
    """Represents a Seasonal Defense Module.

    Attributes
    ----------
    id: :class:`int`
        The module's unique identifier.
    name: :class:`str`
        The module's name.
    TID: :class:`TID`
        The module's translation IDs for localization.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this module.
    max_level: :class:`int`
        The maximum level this module can be upgraded to.
    build_cost: :class:`int`
        The cost to build/upgrade to this level.
    build_time: :class:`TimeDelta`
        The time required to build/upgrade to this level.
    ability_data: :class:`dict`
        The ability data for this module.
    
    Note
    ----
    To get the upgrade cost, access the `build_cost` of the next level.
    """

    __slots__ = (
        "id",
        "name",
        "TID",
        "upgrade_resource",
        "max_level",
        "build_cost",
        "build_time",
        "ability_data",
    )

    def __init__(self, data: dict, level: int = 0):
        super().__init__(
            initial_level=level,
            static_data=data
        )
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.TID: TID = TID(data=data["TID"])
        self.upgrade_resource = Resource(value=data["upgrade_resource"])

        self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.build_cost: int = level_data.get("build_cost")
        self.build_time: TimeDelta = TimeDelta(seconds=level_data.get("build_time"))
        self.ability_data: dict = level_data.get("ability_data")

class SeasonalDefense(BaseDataClass):
    """Represents a Seasonal Defense.

    Attributes
    ----------
    id: :class:`int`
        The defense's unique identifier.
    name: :class:`str`
        The defense's name.
    info: :class:`str`
        Description of the defense.
    TID: :class:`TID`
        The defense's translation IDs for localization.
    modules: List[:class:`SeasonalDefenseModule`]
        The list of modules for this defense.
    level: :class:`int`
        The total level (sum of all module levels).
    """

    __slots__ = (
        "id",
        "name",
        "info",
        "TID",
        "modules",
        "level",
    )

    def __init__(self, data: dict, modules: list[SeasonalDefenseModule]):
        self.id = data["_id"]
        self.name: str = data["name"]
        self.info: str = data["info"]
        self.TID = TID(data=data["TID"])

        self.modules = modules
        self.level = sum(m.level for m in self.modules)

class MergeRequirement:
    """Represents a building merge requirement.

    Attributes
    ----------
    id: :class:`int`
        The building's unique identifier.
    name: :class:`str`
        The building's name.
    geared_up: :class:`bool`
        Whether the building needs to be geared up.
    required_level: :class:`int`
        The required level of the building.
    """

    __slots__ = (
        "id",
        "name",
        "geared_up",
        "required_level",
    )

    def __init__(self, data: dict):
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.geared_up: bool = data["geared_up"]
        self.required_level: int = data["level"]

class GearUp:
    """Represents a gear up requirement for a building.

    Attributes
    ----------
    required_level: :class:`int`
        The required level to gear up.
    building_id: :class:`int`
        The building ID required for gear up.
    resource: :class:`Resource`
        The resource type required for gear up.
    """

    __slots__ = (
        "required_level",
        "building_id",
        "resource",
    )

    def __init__(self, data: dict):
        self.required_level: int = data["level_required"]
        self.building_id: int = data["building_id"]
        self.resource: Resource = Resource(value=data["resource"])

class TownhallUnlock:
    """Represents a building unlock at a townhall level.

    Attributes
    ----------
    name: :class:`str`
        The name of the unlocked item.
    _id: :class:`int`
        The id of the unlocked building.
    quantity: :class:`int`
        The quantity unlocked.
    """

    __slots__ = (
        "name",
        "_id",
        "quantity",
    )

    def __init__(self, data: dict):
        self.name: str = data["name"]
        self._id: int = data["_id"]
        self.quantity: int = data["quantity"]

class TownhallWeapon(LeveledUnit):
    """Represents a Townhall Weapon.

    Attributes
    ----------
    name: :class:`str`
        The weapon's name.
    info: :class:`str`
        Description of the weapon.
    TID: :class:`TID`
        The weapon's translation IDs for localization.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this weapon.
    build_cost: :class:`int`
        The cost to build/upgrade to this level.
    build_time: :class:`TimeDelta`
        The time required to build/upgrade to this level.
    dps: :class:`int`
        The weapon's damage per second.
    
    Note
    ----
    To get the upgrade cost, access the `build_cost` of the next level.
    """

    __slots__ = (
        "name",
        "info",
        "TID",
        "upgrade_resource",
        "build_cost",
        "build_time",
        "dps",
    )

    def __init__(self, data: dict, level: int = 0):
        super().__init__(
            initial_level=level,
            static_data=data
        )
        self.name: str = data["name"]
        self.info: str = data["info"]
        self.TID: TID = TID(data=data["TID"])
        self.upgrade_resource: Resource = Resource(value=data["upgrade_resource"])

        self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.build_cost: int = level_data["build_cost"]
        self.build_time = TimeDelta(seconds=level_data["build_time"])
        self.dps: int = level_data["dps"]

class Supercharge(LevelManager):
    """Represents a Supercharge for a building.

    Attributes
    ----------
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this supercharge.
    build_cost: :class:`int`
        The cost to build/upgrade to this level.
    build_time: :class:`TimeDelta`
        The time required to build/upgrade to this level.
    hitpoints_buff: :class:`int`
        The hitpoints buff provided by this supercharge.
    dps_buff: :class:`int`
        The damage per second buff provided by this supercharge.
    
    Note
    ----
    To get the upgrade cost, access the `build_cost` of the next level.
    """

    __slots__ = (
        "upgrade_resource",
        "build_cost",
        "build_time",
        "hitpoints_buff",
        "dps_buff",
    )

    def __init__(self, data: dict, level: int = 0):
        super().__init__(
            initial_level=level,
            static_data=data
        )
        self.upgrade_resource: Resource = Resource(value=data["upgrade_resource"])

        self._load_level_data()

    def _load_level_data(self):
        if not self._static_data:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.build_cost: int = level_data["build_cost"]
        self.build_time = TimeDelta(seconds=level_data["build_time"])
        self.hitpoints_buff: int = level_data["hitpoints_buff"]
        self.dps_buff: int = level_data["dps_buff"]

class Building(LeveledUnit):
    """Represents a Building.

    Attributes
    ----------
    id: :class:`int`
        The building's unique identifier.
    name: :class:`str`
        The building's name.
    info: :class:`str`
        Description of the building.
    TID: :class:`TID`
        The building's translation IDs for localization.
    type: :class:`BuildingType`
        The type of building.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this building.
    village: :class:`VillageType`
        The village type where this building belongs.
    width: :class:`int`
        The width of the building.
    is_superchargeable: :class:`bool`
        Whether this building can be supercharged.
    gear_up: Optional[:class:`GearUp`]
        The gear up requirements for this building.
    seasonal_defenses: List[:class:`SeasonalDefense`]
        The list of seasonal defenses for this building (Crafting Station only).
    weapon: Optional[:class:`TownhallWeapon`]
        The townhall weapon (only for townhall).
    build_cost: :class:`int`
        The cost to build/upgrade to this level.
    build_time: :class:`TimeDelta`
        The time required to build/upgrade to this level.
    required_townhall: :class:`int`
        The townhall level required to build/upgrade to this level.
    hitpoints: :class:`int`
        The building's hitpoints.
    dps: :class:`int`
        The building's damage per second.
    supercharge: Optional[:class:`Supercharge`]
        The supercharge for this building.
    merge_requirement: List[:class:`MergeRequirement`]
        The merge requirements for this building.
    unlocks: List[:class:`TownhallUnlock`]
        The unlocks provided by this building (only for townhall).
    
    Note
    ----
    To get the upgrade cost, access the `build_cost` of the next level.
    """

    __slots__ = (
        "_supercharge_level",
        "_weapon_level",
        "id",
        "name",
        "info",
        "TID",
        "type",
        "upgrade_resource",
        "village",
        "width",
        "is_superchargeable",
        "gear_up",
        "seasonal_defenses",
        "weapon",
        "build_cost",
        "build_time",
        "required_townhall",
        "hitpoints",
        "dps",
        "supercharge",
        "merge_requirement",
        "unlocks",
    )

    def __init__(self,
         data: dict, level: int = 0, weapon_level: int = 0,
         supercharge_level: int = 0,
         seasonal_defenses: list[SeasonalDefense] = None
    ):
        super().__init__(
            initial_level=level,
            static_data=data
        )
        self._supercharge_level = supercharge_level
        self._weapon_level = weapon_level

        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.info: str = data["info"]
        self.TID: TID = TID(data=data["TID"])
        self.type = BuildingType(value=data["type"])
        self.upgrade_resource: Resource = Resource(value=data["upgrade_resource"])
        self.village = VillageType(value=data["village"])
        self.width: int = data["width"]
        self.is_superchargeable: bool = data["superchargeable"]

        # only some buildings
        self.gear_up = None
        if "gear_up" in data:
            self.gear_up = GearUp(data=data["gear_up"])

        self.seasonal_defenses = seasonal_defenses or []
        if "seasonal_defenses" in data and not self.seasonal_defenses:
            # if no seasonal defenses are provided, initialize a set
            seasonal_defenses = []
            for defense_data in data["seasonal_defenses"]:
                modules = []
                for module_data in defense_data["modules"]:
                    modules.append(SeasonalDefenseModule(data=module_data))
                seasonal_defenses.append(SeasonalDefense(data=defense_data, modules=modules))
            self.seasonal_defenses = seasonal_defenses

        self._load_level_data()

    def _load_level_data(self):
        if not self._static_data or not self._static_data["levels"]:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.weapon: TownhallWeapon | None = None
        if "weapon" in level_data:
            self.weapon = TownhallWeapon(level=self._weapon_level, data=level_data["weapon"])

        self.build_cost: int = level_data["build_cost"]
        self.build_time = TimeDelta(seconds=level_data["build_time"])
        self.required_townhall: int = level_data["required_townhall"]
        self.hitpoints: int = level_data["hitpoints"]
        self.dps: int = level_data["dps"]

        self.supercharge: Supercharge | None = None
        if "supercharge" in level_data:
            self.supercharge = Supercharge(level=self._supercharge_level, data=level_data["supercharge"])
        # only select buildings
        self.merge_requirement = [MergeRequirement(m) for m in level_data.get("merge_requirement", [])]

        # only the townhall
        self.unlocks = [TownhallUnlock(u) for u in level_data.get("unlocks", [])]

class Trap(LeveledUnit):
    """Represents a Trap.

    Attributes
    ----------
    id: :class:`int`
        The trap's unique identifier.
    name: :class:`str`
        The trap's name.
    TID: :class:`TID`
        The trap's translation IDs for localization.
    upgrade_resource: :class:`Resource`
        The resource type required to upgrade this trap.
    village: :class:`VillageType`
        The village type where this trap belongs.
    width: :class:`int`
        The width of the trap.
    is_air_triggerable: :class:`bool`
        Whether this trap can be triggered by air units.
    is_ground_triggerable: :class:`bool`
        Whether this trap can be triggered by ground units.
    damage_radius: :class:`int`
        The damage radius of the trap.
    trigger_radius: :class:`int`
        The trigger radius of the trap.
    build_cost: :class:`int`
        The cost to build/upgrade to this level.
    build_time: :class:`TimeDelta`
        The time required to build/upgrade to this level.
    required_townhall: :class:`int`
        The townhall level required to build/upgrade to this level.
    damage: :class:`int`
        The damage dealt by the trap.
    
    Note
    ----
    To get the upgrade cost, access the `build_cost` of the next level.
    """

    __slots__ = (
        "id",
        "name",
        "TID",
        "upgrade_resource",
        "village",
        "width",
        "is_air_triggerable",
        "is_ground_triggerable",
        "damage_radius",
        "trigger_radius",
        "build_cost",
        "build_time",
        "required_townhall",
        "damage",
    )

    def __init__(self, data: dict, level: int = 0):
        super().__init__(
            initial_level=level,
            static_data=data
        )
        self.id: int = data["_id"]
        self.name: str = data["name"]
        self.TID: TID = TID(data=data["TID"])

        self.upgrade_resource: Resource = Resource(value=data["upgrade_resource"])
        self.village = VillageType(value=data["village"])
        self.width: int = data["width"]
        self.is_air_triggerable: bool = data["air_trigger"]
        self.is_ground_triggerable: bool = data["ground_trigger"]
        self.damage_radius: int = data["damage_radius"]
        self.trigger_radius: int = data["trigger_radius"]


    def _load_level_data(self):
        if not self._static_data or not self._static_data["levels"]:
            return

        level_data = self._static_data["levels"][self._level - 1]

        self.build_cost: int = level_data["build_cost"]
        self.build_time = TimeDelta(seconds=level_data["build_time"])
        self.required_townhall: int = level_data["required_townhall"]
        self.damage: int = level_data["damage"]


from typing import Iterable, Callable, Union, Coroutine, Type, Dict

from coc.client import Client
from coc.players import Player, ClanMember
from coc.clans import Clan
from coc.wars import ClanWar
from coc.war_attack import WarAttack

_ClanType = Type[Clan]
_PlayerType = Type[Player]
_CustomClassType = Union[Type[Clan], Type[Player], Type[ClanWar]]
_EventPredicateClasses = Union[Clan, Player, ClanWar, ClanMember, WarAttack]
_EventCallbackType = Callable[[_EventPredicateClasses, _EventPredicateClasses], Coroutine]
_EventCallbackCustomArgumentsType = Callable[..., Coroutine]
_EventPredicateType = Callable[[_EventPredicateClasses, _EventPredicateClasses], bool]
_EventWrappedPredicateType = Callable[
    [_EventPredicateClasses, _EventPredicateClasses, _EventCallbackType], Coroutine[None]
]
_EventDecoratorReturn = Callable[[_EventCallbackType], _EventCallbackType]

class ClanEvents:
    event_type: str
    @classmethod
    def name(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def badge(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def level(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def points(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def versus_points(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...

    @classmethod
    def builder_base_points(cls,
                            tags: Iterable = None,
                            custom_class: _ClanType = Clan,
                            retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_count(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def location(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def type(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def required_trophies(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def war_frequency(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def war_win_streak(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def war_wins(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def war_ties(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def war_losses(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def public_war_log(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def description(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def war_league(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def labels(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_join(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_leave(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_name(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_role(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_exp_level(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_league(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_trophies(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_clan_rank(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_clan_previous_rank(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_versus_trophies(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_builder_base_trophies(cls,
                                     tags: Iterable = None,
                                     custom_class: _ClanType = Clan,
                                     retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_versus_rank(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...

    @classmethod
    def member_builder_base_rank(cls,
                                 tags: Iterable = None,
                                 custom_class: _ClanType = Clan,
                                 retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_donations(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def member_received(cls, tags: Iterable = None, custom_class: _ClanType = Clan, retry_interval: int = None) -> _EventDecoratorReturn: ...

class PlayerEvents:
    event_type: str
    @classmethod
    def achievement_change(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def name(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def exp_level(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def trophies(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def versus_trophies(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...

    @classmethod
    def builder_base_trophies(cls,
                              tags: Iterable = None,
                              custom_class: _PlayerType = Player,
                              retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def clan_rank(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def clan_previous_rank(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def donations(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def received(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def league(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def role(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def best_trophies(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def war_stars(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def town_hall(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def town_hall_weapon(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def builder_hall(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def best_versus_trophies(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def versus_attack_wins(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def best_builder_base_trophies(cls,
                                   tags: Iterable = None,
                                   custom_class: _PlayerType = Player,
                                   retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def legend_statistics(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def labels(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def equipment_change(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def active_equipment_change(cls, tags: Iterable = None, custom_class: _PlayerType = Player, retry_interval: int = None) -> _EventDecoratorReturn: ...

class WarEvents:
    event_type: str
    @classmethod
    def war_attack(cls, tags: Iterable = None, custom_class: Type[ClanWar] = ClanWar, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def state(cls, tags: Iterable = None, custom_class: Type[ClanWar] = ClanWar, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def members(cls, tags: Iterable = None, custom_class: Type[ClanWar] = ClanWar, retry_interval: int = None) -> _EventDecoratorReturn: ...
    @classmethod
    def new_war(cls, tags: Iterable = None, custom_class: Type[ClanWar] = ClanWar, retry_interval: int = None) -> _EventDecoratorReturn: ...


class ClientEvents:
    @classmethod
    def maintenance_start(cls): ...
    @classmethod
    def maintenance_completion(cls): ...
    @classmethod
    def new_season_start(cls): ...
    @classmethod
    def raid_weekend_start(cls): ...
    @classmethod
    def raid_weekend_end(cls): ...
    @classmethod
    def clan_games_start(cls): ...
    @classmethod
    def clan_games_end(cls): ...
    @classmethod
    def keys_reset(cls): ...
    @classmethod
    def event_error(cls): ...
    @classmethod
    def clan_loop_start(cls): ...
    @classmethod
    def clan_loop_finish(cls): ...
    @classmethod
    def player_loop_start(cls): ...
    @classmethod
    def player_loop_finish(cls): ...
    @classmethod
    def war_loop_start(cls): ...
    @classmethod
    def war_loop_finish(cls): ...

class EventsClient(Client):
    clan_retry_interval: int
    player_retry_interval: int
    war_retry_interval: int

    clan_cls: Type[Clan]
    player_cls: Type[Player]
    war_cls: Type[ClanWar]

    _player_updates: set
    _clan_updates: set
    _war_updates: set

    _listeners: Dict


    is_cwl_active: bool

    def add_clan_updates(self, *tags: str) -> None: ...
    def add_player_updates(self, *tags: str) -> None: ...
    def add_war_updates(self, *tags: str) -> None: ...
    def remove_clan_updates(self, *tags: str) -> None: ...
    def remove_player_updates(self, *tags: str) -> None: ...
    def remove_war_updates(self, *tags: str) -> None: ...
    def event(self, event: Callable) -> Callable: ...
    def add_events(self, *events: Callable) -> None: ...
    def remove_events(self, *events: Callable) -> None: ...
    def run_forever(self) -> None: ...
